import re
import os
import papis.downloaders.base
import arxiv2bib


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="arxiv")
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url):
        m = re.match(r"^arxiv:(.*)", url, re.IGNORECASE)
        if m:
            url = "https://arxiv.org/abs/{m}".format(m=m.group(1))
        if re.match(r".*arxiv.org.*", url):
            # https://arxiv.org/pdf/1707.09820|.pdf?blahslas=sdfad|
            url = re.sub(r"\.pdf.*$", "", url)
            return Downloader(url)
        else:
            return False

    def get_identifier(self):
        """Get arxiv identifier
        :returns: Identifier
        """
        url = self.get_url()
        return re.sub(r'^.*arxiv.org/(abs|pdf)/(.*)\/?$', r'\2', url)

    def get_bibtex_url(self):
        identifier = self.get_identifier()
        return identifier

    def download_bibtex(self):
        bib_url = self.get_bibtex_url()
        bibtexCli = arxiv2bib.Cli([bib_url])
        bibtexCli.run()
        self.logger.debug("[bibtex url] = %s" % bib_url)
        data = os.linesep.join(bibtexCli.output)
        self.bibtex_data = data

    def get_document_url(self):
        # https://arxiv.org/pdf/1702.01590
        identifier = self.get_identifier()
        self.logger.debug("[paper id] = %s" % identifier)
        pdf_url = "https://arxiv.org/pdf/"+identifier+".pdf"
        self.logger.debug("[pdf url] = %s" % pdf_url)
        return pdf_url



"""

Cli
^^^
.. click:: papis.commands.rm:cli
    :prog: papis rm
"""
import papis
import papis.api
from papis.api import status
import papis.utils
import papis.config
import papis.document
import papis.cli
import papis.database
import click


def run(
        document,
        filepath=None
        ):
    """Main method to the rm command

    :returns: List different objects
    :rtype:  list
    """
    db = papis.database.get()
    if filepath is not None:
        document.rm_file(filepath)
        document.save()
    else:
        papis.document.delete(document)
        db.delete(document)
    return status.success


@click.command()
@click.help_option('-h', '--help')
@papis.cli.query_option()
@click.option(
    "--file",
    help="Remove files from a document instead of the whole folder",
    is_flag=True,
    default=False
)
@click.option(
    "-f", "--force",
    help="Do not confirm removal",
    is_flag=True,
    default=False
)
def cli(
        query,
        file,
        force
        ):
    """Delete command for several objects"""
    documents = papis.database.get().query(query)
    document = papis.api.pick_doc(documents)
    if not document:
        return status.file_not_found
    if file:
        filepath = papis.api.pick(
            document.get_files()
        )
        if not filepath:
            return status.file_not_found
        if not force:
            toolbar = 'The file {0} would be removed'.format(filepath)
            if not papis.utils.confirm("Are you sure?", bottom_toolbar=toolbar):
                return status.success
        click.echo("Removing %s..." % filepath)
        return run(
            document,
            filepath=filepath
        )
    else:
        if not force:
            toolbar = 'The folder {0} would be removed'.format(
                document.get_main_folder()
            )
            if not papis.utils.confirm("Are you sure?", bottom_toolbar=toolbar):
                return status.success
        click.echo("Removing ...")
        return run(document)



"""This command is to update the information of the documents.
Some examples of the usage are given below

Examples
^^^^^^^^

- Update a document automatically and interactively
  (searching by ``doi`` number in *crossref*, or in other sources...)

    .. code::

        papis update --auto -i "author = dyson"

- Update your library from a bib(la)tex file where many entries are listed.
  papis will try to look for documents in your library that match these
  entries and will ask you entry per entry to update it (of course this is
  done if you use the ``-i`` flag for interactively doing it). In the example
  ``libraryfile.bib`` is a file containing many entries.

    .. code::

        papis update --from-bibtex libraryfile.bib -i

- Tag all einstein papers with the tag classics

    .. code::

        papis update --all --set tags classics einstein

and add the tag of ``physics`` to all papers tagged as ``classics``

    .. code::

        papis update --all --set tags '{doc[tags]} physics' einstein

Cli
^^^
.. click:: papis.commands.update:cli
    :prog: papis update
"""

import papis
import urllib.error
import logging
import papis.utils
import papis.bibtex
import papis.downloaders.utils
import papis.document
import papis.database
import papis.isbnplus
import papis.isbn
import papis.crossref
import papis.api
import papis.cli
import click
import yaml


def run(document, data=dict(), interactive=False, force=False):
    # Keep the ref the same, otherwise issues can be caused when
    # writing LaTeX documents and all the ref's change
    data['ref'] = document['ref']

    document.update(data, force, interactive)
    document.save()
    papis.database.get().update(document)


@click.command()
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    "-i/-b",
    "--interactive/--no-interactive",
    help="Interactively update",
    default=True
)
@click.option(
    "-f",
    "--force",
    help="Force update, overwrite conflicting information",
    default=False,
    is_flag=True
)
@click.option(
    "--from-crossref",
    help="Update info from crossref.org",
    default=None
)
@click.option(
    "--from-isbn",
    help="Update info from isbn",
    default=None
)
@click.option(
    "--from-isbnplus",
    help="Update info from isbnplus.org",
    default=None
)
@click.option(
    "--from-yaml",
    help="Update info from yaml file",
    default=None
)
@click.option(
    "--from-bibtex",
    help="Update info from bibtex file",
    default=None
)
@click.option(
    "--from-url",
    help="Get document or information from url",
    default=None
)
@click.option(
    "--from-doi",
    help="Doi to try to get information from",
    default=None
)
@click.option(
    "--auto",
    help="Try to parse information from different sources",
    default=False,
    is_flag=True
)
@click.option(
    "--all",
    help="Update all entries in library",
    default=False,
    is_flag=True
)
@click.option(
    "-s", "--set",
    help="Update document's information with key value."
         "The value can be a papis format.",
    multiple=True,
    type=(str, str),
)
def cli(
        query,
        interactive,
        force,
        from_crossref,
        from_isbnplus,
        from_isbn,
        from_yaml,
        from_bibtex,
        from_url,
        from_doi,
        auto,
        all,
        set
        ):
    """Update a document from a given library"""

    documents = papis.database.get().query(query)
    data = dict()
    logger = logging.getLogger('cli:update')

    if not all:
        documents = [papis.api.pick_doc(documents)]

    for document in documents:
        if all:
            data = dict()
            from_url = None
            from_doi = None
            from_isbnplus = None
            from_isbnplus = None

        if set:
            data.update(
                {s[0]: papis.utils.format_doc(s[1], document) for s in set}
            )

        if auto:
            if 'doi' in document.keys() and not from_doi:
                logger.info('Trying using the doi {}'.format(document['doi']))
                from_doi = document['doi']
            if 'url' in document.keys() and not from_url:
                logger.info('Trying using the url {}'.format(document['url']))
                from_url = document['url']
            if 'title' in document.keys() and not from_isbn:
                logger.info(
                    'Trying using the title {}'.format(document['title'])
                )
                from_isbn = '{d[title]} {d[author]}'.format(d=document)
                from_isbnplus = from_isbn
            if from_crossref is None and from_doi is None:
                from_crossref = True

        if from_crossref:
            if from_crossref is True:
                from_crossref = ''
            try:
                doc = papis.api.pick_doc([
                        papis.document.from_data(d)
                        for d in papis.crossref.get_data(
                            query=from_crossref,
                            author=document['author'],
                            title=document['title']
                        )
                ])
                if doc:
                    data.update(papis.document.to_dict(doc))
                    if 'doi' in document.keys() and not from_doi and auto:
                        from_doi = doc['doi']

            except Exception as e:
                logger.error(e)

        if from_isbnplus:
            try:
                doc = papis.api.pick_doc(
                    [
                        papis.document.from_data(d)
                        for d in papis.isbnplus.get_data(
                            query=from_isbnplus
                        )
                    ]
                )
                if doc:
                    data.update(papis.document.to_dict(doc))
            except urllib.error.HTTPError:
                logger.error('urllib failed to download')

        if from_isbn:
            try:
                doc = papis.api.pick_doc(
                    [
                        papis.document.from_data(d)
                        for d in papis.isbn.get_data(
                            query=from_isbn
                        )
                    ]
                )
                if doc:
                    data.update(papis.document.to_dict(doc))
            except Exception as e:
                logger.error('Isbnlib had an error retrieving information')
                logger.error(e)

        if from_yaml:
            with open(from_yaml) as fd:
                data.update(yaml.load(fd))

        if from_doi:
            logger.debug("Try using doi %s" % from_doi)
            data.update(papis.utils.doi_to_data(from_doi))

        if from_bibtex:
            try:
                bib_data = papis.bibtex.bibtex_to_dict(from_bibtex)
                data.update(bib_data[0])
            except Exception:
                pass

        if from_url:
            try:
                url_data = papis.downloaders.utils.get(from_url)
                data.update(url_data["data"])
            except Exception:
                pass

        run(
            document,
            data=data,
            interactive=interactive,
            force=force,
        )



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="acs")
        self.expected_document_extension = 'pdf'
        # It seems to be necessary so that acs lets us download the bibtex
        self.cookies = {
            'gdpr': 'true',
        }

    @classmethod
    def match(cls, url):
        if re.match(r".*acs.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_doi(self):
        # .../pubs.acs.org/doi/whatever/10.1021/acs.jchemed.6b00559?blah=2
        mdoi = re.match(r'.*acs.org/doi/[^/]+/([^?]*)', self.get_url())
        if mdoi:
            doi = mdoi.group(1).replace("abs/", "").replace("full/", "")
            self.logger.debug("[doi] = %s" % doi)
            return doi
        else:
            self.logger.error("Doi not found!!")

    def get_document_url(self):
        # http://pubs.acs.org/doi/pdf/10.1021/acs.jchemed.6b00559
        return "http://pubs.acs.org/doi/pdf/" + self.get_doi()

    def get_bibtex_url(self):
        url = "http://pubs.acs.org/action/downloadCitation"\
              "?format=bibtex&cookieSet=1&doi=%s" % self.get_doi()
        self.logger.debug("[bibtex url] = %s" % url)
        return url



import click
import click.decorators
import os
import papis.config
import glob
import re


default_commands_path = os.path.join(os.path.dirname(__file__), 'commands')


def get_external_scripts():
    paths = []
    scripts = []
    paths.append(papis.config.get_scripts_folder())
    paths += os.environ["PATH"].split(":")
    for path in paths:
        scripts += glob.glob(os.path.join(path, "papis-*"))
    return scripts


def script_paths_to_dict(paths):
    scripts_dict = dict()
    for path in paths:
        command_name = \
            os.path.basename(path).replace('papis-', '').replace('.py', '')
        scripts_dict[command_name] = dict(
            command_name=command_name,
            path=path,
            python=path.endswith(".py")
        )
    return scripts_dict


class MultiCommand(click.MultiCommand):

    scripts = script_paths_to_dict(
        list(filter(
            lambda path: not re.match(r'.*__init__.py$', path),
            glob.glob(os.path.join(default_commands_path, '*.py'))
        )) +
        get_external_scripts()
    )

    def list_commands(self, ctx):
        """List all matched commands in the command folder and in path

        >>> mc = MultiCommand()
        >>> rv = mc.list_commands(None)
        >>> len(rv) > 0
        True
        """
        rv = [s for s in self.scripts.keys()]
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Get the command to be run

        >>> mc = MultiCommand()
        >>> cmd = mc.get_command(None, 'add')
        >>> cmd.name, cmd.help
        ('cli', 'Add...')
        >>> mc.get_command(None, 'this command does not exist')
        """
        namespace = {}
        try:
            script = self.scripts[name]
        except KeyError:
            return None
        if script['python']:
            with open(script['path']) as f:
                code = compile(f.read(), script['path'], 'exec')
                eval(code, namespace, namespace)
            if 'cli' in namespace.keys():
                return namespace['cli']
        # If it gets here, it means that it is an external script
        from papis.commands.external import external_cli as cli
        from papis.commands.external import get_command_help
        cli.context_settings['obj'] = script
        cli.help = get_command_help(script['path'])
        cli.short_help = cli.help
        return cli


def query_option(**attrs):
    """Adds a ``query`` argument as a decorator"""
    def decorator(f):
        import papis.config
        attrs.setdefault(
            'default',
            lambda: papis.config.get('default-query-string')
        )
        return click.decorators.argument('query', **attrs)(f)
    return decorator


def git_option(**attrs):
    """Adds a ``git`` option as a decorator"""
    def decorator(f):
        import papis.config
        attrs.setdefault(
            'default',
            lambda: True if papis.config.get('use-git') else False
        )
        attrs.setdefault('help', 'Add git interoperability')
        return click.decorators.option('--git/--no-git', **attrs)(f)
    return decorator



"""The following table lists the field prefixes for all the fields
 that can be searched.

 Table:          search_query field prefixes
============================================
 prefix          explanation
--------------------------------------------
 ti              Title
 au              Author
 abs             Abstract
 co              Comment
 jr              Journal Reference
 cat             Subject Category
 rn              Report Number
 id              Id (use id_list instead)
 all             All of the above
"""
import bs4
import logging
import urllib.request  # urlopen, Request
import urllib.parse  # import urlencode
import papis.config


logger = logging.getLogger('arxiv')


def get_data(
    query="",
    author="",
    title="",
    abstract="",
    comment="",
    journal="",
    report_number="",
    category="",
    id_list="",
    page=0,
    max_results=30
):
    dict_params = {
        "all": query,
        "ti": title,
        "au": author,
        "cat": category,
        "abs": abstract,
        "co": comment,
        "jr": journal,
        "id_list": id_list,
        "rn": report_number
    }
    result = []
    clean_params = {x: dict_params[x] for x in dict_params if dict_params[x]}
    search_query = '+AND+'.join(
        [key+':'+str(clean_params[key]) for key in clean_params]
    )
    logger.debug("query = " + search_query)
    params = urllib.parse.urlencode(
        {
            'search_query': search_query,
            'start': page,
            'max_results': max_results
        }
    )
    main_url = "http://arxiv.org/api/query?"
    req_url = main_url + params
    logger.debug("url = " + req_url)
    url = urllib.request.Request(
        req_url,
        headers={
            'User-Agent': papis.config.get('user-agent')
        }
    )
    xmldoc = urllib.request.urlopen(url).read()
    soup = bs4.BeautifulSoup(xmldoc, 'html.parser')

    entries = soup.find_all("entry")
    for entry in entries:
        data = dict()
        data["abstract"] = entry.find("summary").get_text().replace(
            "\n", " "
        )
        data["url"] = entry.find("id").get_text()
        data["published"] = entry.find("published").get_text()
        data["year"] = data.get("published")[0:4]
        data["title"] = entry.find("title").get_text().replace("\n", " ")
        data["author"] = ", ".join(
            [
                author.get_text().replace("\n", "")
                for author in entry.find_all("author")
            ]
        )
        result.append(data)
    return result



import logging
import os
import papis.utils
import papis.docmatcher
import papis.config
import papis.database.base
import re


logger = logging.getLogger("cache")


def get_cache_home():
    """Get folder where the cache files are stored, it retrieves the
    ``cache-dir`` configuration setting. It is ``XDG`` standard compatible.

    :returns: Full path for cache main folder
    :rtype:  str

    >>> import os; os.environ["XDG_CACHE_HOME"] = '~/.cache'
    >>> get_cache_home() == os.path.expanduser(\
            os.path.join(os.environ["XDG_CACHE_HOME"], 'papis')\
        )
    True
    >>> os.environ["XDG_CACHE_HOME"] = '/tmp/.cache'
    >>> get_cache_home()
    '/tmp/.cache/papis'
    >>> del os.environ["XDG_CACHE_HOME"]
    >>> get_cache_home() == os.path.expanduser(\
            os.path.join('~/.cache', 'papis')\
        )
    True
    """
    user_defined = papis.config.get('cache-dir')
    if user_defined is not None:
        return os.path.expanduser(user_defined)
    else:
        return os.path.expanduser(
            os.path.join(os.environ.get('XDG_CACHE_HOME'), 'papis')
        ) if os.environ.get(
            'XDG_CACHE_HOME'
        ) else os.path.expanduser(
            os.path.join('~', '.cache', 'papis')
        )


def get(path):
    """Get contents stored in a cache file ``path`` in pickle binary format.

    :param path: Path to the cache file.
    :type  path: str
    :returns: Content of the cache file.
    :rtype: object

    >>> create([1,2,3], '/tmp/test-pickle')
    >>> get('/tmp/test-pickle')
    [1, 2, 3]
    """
    import pickle
    logger.debug("Getting cache %s " % path)
    return pickle.load(open(path, "rb"))


def create(obj, path):
    """Create a cache file in ``path`` with obj as its content using pickle
    binary format.

    :param obj: Any seriazable object.
    :type  obj: object
    :param path: Path to the cache file.
    :type  path: str
    :returns: Nothing
    :rtype: None
    """
    import pickle
    logger.debug("Saving in cache %s " % path)
    pickle.dump(obj, open(path, "wb+"))


def get_name(directory):
    """Create a cache file name out of the path of a given directory.

    :param directory: Folder name to be used as a seed for the cache name.
    :type  directory: str
    :returns: Name for the cache file.
    :rtype:  str

    >>> get_name('path/to/my/lib')
    'a8c689820a94babec20c5d6269c7d488-lib'
    >>> get_name('papers')
    'a566b2bebc62611dff4cdaceac1a7bbd-papers'
    """
    import hashlib
    return "{}-{}".format(
        hashlib.md5(directory.encode()).hexdigest(),
        os.path.basename(directory)
    )


def clear(directory):
    """Clear cache associated with a directory

    :param directory: Folder name that was used as a seed for the cache name.
    :type  directory: str
    :returns: Nothing
    :rtype: None

    >>> create([1,2,3], get_cache_file_path('some/other/papers'))
    >>> clear('some/other/papers')
    >>> import os; os.path.exists(get_cache_file_path('some/other/papers'))
    False
    >>> clear('non/existing/some/other/books')
    >>> os.path.exists(get_cache_file_path('non/existing/some/other/books'))
    False
    """
    directory = os.path.expanduser(directory)
    cache_path = get_cache_file_path(directory)
    if os.path.exists(cache_path):
        logger.debug("Clearing cache %s " % cache_path)
        os.remove(cache_path)


def clear_lib_cache(lib=None):
    """Clear cache associated with a library. If no library is given
    then the current library is used.

    :param lib: Library name.
    :type  lib: str

    >>> import os
    >>> if not os.path.exists('/tmp/setlib-test'): os.makedirs(\
            '/tmp/setlib-test'\
        )
    >>> papis.config.set_lib('/tmp/setlib-test')
    >>> create([1,2,3], get_cache_file_path('/tmp/setlib-test'))
    >>> os.path.exists(get_cache_file_path('/tmp/setlib-test'))
    True
    >>> clear_lib_cache('/tmp/setlib-test')
    >>> os.path.exists(get_cache_file_path('/tmp/setlib-test'))
    False
    """
    directory = papis.config.get("dir", section=lib)
    clear(directory)


def get_cache_file_path(directory):
    """Get the full path to the cache file

    :param directory: Library folder
    :type  directory: str

    >>> import os; os.environ["XDG_CACHE_HOME"] = '/tmp'
    >>> get_cache_file_path('blah/papers')
    '/tmp/papis/c39177eca0eaea2e21134b0bd06631b6-papers'
    """
    cache_name = get_name(directory)
    return os.path.expanduser(os.path.join(get_cache_home(), cache_name))


def get_folders(directory):
    """Get folders from within a containing folder from cache

    :param directory: Folder to look for documents (or library folder)
    :type  directory: str
    :param search: Valid papis search
    :type  search: str
    :returns: List of document objects.
    :rtype: list

    >>> create(['/a', '/b'], get_cache_file_path('/asdf/papers'))
    >>> get_folders('/asdf/papers')
    ['/a', '/b']

    """
    cache_home = get_cache_home()
    cache_path = get_cache_file_path(directory)
    folders = []
    logger.debug("Getting documents from dir %s" % directory)
    logger.debug("Cache path = %s" % cache_path)
    if not os.path.exists(cache_home):
        logger.debug("Creating cache dir %s " % cache_home)
        os.makedirs(cache_home, mode=papis.config.getint('dir-umask'))
    if os.path.exists(cache_path):
        logger.debug("Loading folders from cache")
        folders = get(cache_path)
    else:
        folders = papis.utils.get_folders(directory)
        create(folders, cache_path)
    return folders


def filter_documents(documents, search=""):
    """Filter documents. It can be done in a multi core way.

    :param documents: List of papis documents.
    :type  documents: papis.documents.Document
    :param search: Valid papis search string.
    :type  search: str
    :returns: List of filtered documents
    :rtype:  list

    >>> document = papis.document.from_data({'author': 'einstein'})
    >>> len(filter_documents([document], search="einstein")) == 1
    True
    >>> len(filter_documents([document], search="author = ein")) == 1
    True
    >>> len(filter_documents([document], search="title = ein")) == 1
    False

    """
    logger = logging.getLogger('filter')
    papis.docmatcher.DocMatcher.set_search(search)
    papis.docmatcher.DocMatcher.parse()
    papis.docmatcher.DocMatcher.set_matcher(match_document)
    if search == "" or search == ".":
        return documents
    else:
        # Doing this multiprocessing in filtering does not seem
        # to help much, I don't know if it's because I'm doing something
        # wrong or it is really like this.
        import multiprocessing
        import time
        np = papis.api.get_arg("cores", multiprocessing.cpu_count())
        pool = multiprocessing.Pool(np)
        logger.debug(
            "Filtering docs (search %s) using %s cores" % (
                search,
                np
            )
        )
        logger.debug("pool started")
        begin_t = time.time()
        result = pool.map(
            papis.docmatcher.DocMatcher.return_if_match, documents
        )
        pool.close()
        pool.join()
        logger.debug("pool done (%s ms)" % (1000*time.time()-1000*begin_t))
        return [d for d in result if d is not None]


def folders_to_documents(folders):
    """Turn folders into documents, this is done in a multiprocessing way, this
    step is quite critical for performance.

    :param folders: List of folder paths.
    :type  folders: list
    :returns: List of document objects.
    :rtype:  list
    """
    import multiprocessing
    import time
    logger = logging.getLogger("dir2doc")
    np = papis.api.get_arg("cores", multiprocessing.cpu_count())
    logger.debug("Running in %s cores" % np)
    pool = multiprocessing.Pool(np)
    logger.debug("pool started")
    begin_t = time.time()
    result = pool.map(papis.document.from_folder, folders)
    pool.close()
    pool.join()
    logger.debug("pool done (%s ms)" % (1000*time.time()-1000*begin_t))
    return result


def match_document(document, search, match_format=None):
    """Main function to match document to a given search.

    :param document: Papis document
    :type  document: papis.document.Document
    :param search: A valid search string
    :type  search: str
    :param match_format: Python-like format string.
        (`see <
            https://docs.python.org/2/library/string.html#format-string-syntax
        >`_)
    :type  match_format: str
    :returns: Non false if matches, true-ish if it does match.

    >>> papis.config.set('match-format', '{doc[author]}')
    >>> document = papis.document.from_data({'author': 'einstein'})
    >>> match_document(document, 'e in') is None
    False
    >>> match_document(document, 'ee in') is None
    True
    >>> match_document(document, 'einstein', '{doc[title]}') is None
    True
    """
    match_format = match_format or papis.config.get("match-format")
    match_string = papis.utils.format_doc(match_format, document)
    regex = get_regex_from_search(search)
    return re.match(regex, match_string, re.IGNORECASE)


def get_regex_from_search(search):
    """Creates a default regex from a search string.

    :param search: A valid search string
    :type  search: str
    :returns: Regular expression
    :rtype: str

    >>> get_regex_from_search(' ein 192     photon')
    '.*.*ein.*192.*photon'
    """
    return r".*"+re.sub(r"\s+", ".*", search)


class Database(papis.database.base.Database):

    def __init__(self, library=None):
        papis.database.base.Database.__init__(self, library)
        self.logger = logging.getLogger('db:cache')
        self.logger.debug('Initializing')
        self.documents = []
        self.folders = []
        self.initialize()

    def get_backend_name(self):
        return 'papis'

    def initialize(self):
        self.get_documents()

    def get_documents(self):
        directory = os.path.expanduser(self.get_dir())

        if papis.config.getboolean("use-cache"):
            self.folders = get_folders(directory)
            self.logger.debug(
                "Loaded folders from cache ({} documents)".format(
                    len(self.folders)
                )
            )
        else:
            self.folders = papis.utils.get_folders(directory)

        self.logger.debug("Creating document objects")
        self.documents = folders_to_documents(self.folders)
        self.logger.debug("Done")

    def add(self, document):
        self.logger.debug('Adding ...')
        self.folders.append(document.get_main_folder())
        assert(self.folders[-1] == document.get_main_folder())
        assert(os.path.exists(document.get_main_folder()))
        self.documents.append(document)
        self.save()

    def update(self, document):
        self.logger.debug('Updating document')

    def delete(self, document):
        if papis.config.getboolean("use-cache"):
            self.logger.debug(
                'Deleting ... ({} documents)'.format(len(self.folders))
            )
            self.folders.remove(document.get_main_folder())
            self.save()
            # Also update the documents list
            self.get_documents()

    def match(self, document, query_string):
        return match_document(document, query_string)

    def clear(self):
        self.logger.debug('Clearing library')
        clear_lib_cache(self.get_lib())

    def query_dict(self, dictionary):
        query_string = " ".join(
            ["{}=\"{}\" ".format(key, val) for key, val in dictionary.items()]
        )
        return self.query(query_string)

    def query(self, query_string):
        self.logger.debug('Querying')
        if len(self.documents) == 0:
            self.get_documents()
        return filter_documents(self.documents, query_string)

    def get_all_query_string(self):
        return '.'

    def get_all_documents(self):
        return self.query(self.get_all_query_string())

    def save(self):
        self.logger.debug(
            'Saving ... ({} documents)'.format(len(self.folders))
        )
        create(self.folders, get_cache_file_path(self.get_dir()))



import os
import shutil
import papis.utils
import papis.config
import papis.bibtex
import logging
import re


logger = logging.getLogger("document")


def open_in_browser(document):
    """Browse document's url whenever possible.

    :document: Document object
    :returns: Returns the url that is composed from the document
    :rtype:  str

    >>> import papis.config; papis.config.set('browser', 'echo')
    >>> papis.config.set('browse-key', 'url')
    >>> open_in_browser( from_data({'url': 'hello.com'}) )
    'hello.com'
    >>> papis.config.set('browse-key', 'doi')
    >>> open_in_browser( from_data({'doi': '12312/1231'}) )
    'https://doi.org/12312/1231'
    >>> papis.config.set('browse-key', 'nonexistentkey')
    >>> open_in_browser( from_data({'title': 'blih', 'author': 'me'}) )
    'https://duckduckgo.com/?q=blih+me'
    """
    global logger
    url = None
    key = papis.config.get("browse-key")

    if document.has(key):
        if "doi" == key:
            url = 'https://doi.org/{}'.format(document['doi'])
        elif "isbn" == key:
            url = 'https://isbnsearch.org/isbn/{}'.format(document['isbn'])
        else:
            url = document[key]

    if url is None or key == 'search-engine':
        from urllib.parse import urlencode
        params = {
            'q': papis.utils.format_doc(
                papis.config.get('browse-query-format'),
                document
            )
        }
        url = papis.config.get('search-engine') + '/?' + urlencode(params)

    logger.debug("Opening url %s:" % url)
    papis.utils.general_open(
        url, "browser", wait=False
    )
    return url


def from_folder(folder_path):
    """Construct a document object from a folder

    :param folder_path: Full path to a valid papis folder
    :type  folder_path: str
    :returns: A papis document
    :rtype:  papis.document.Document
    """
    return papis.document.Document(folder=folder_path)


def from_data(data):
    """Construct a document object from a data dictionary.

    :param data: Data to be copied to a new document
    :type  data: dict
    :returns: A papis document
    :rtype:  papis.document.Document
    """
    return papis.document.Document(data=data)


def to_bibtex(document):
    """Create a bibtex string from document's information

    :param document: Papis document
    :type  document: Document
    :returns: String containing bibtex formating
    :rtype:  str

    >>> import papis.config
    >>> papis.config.set('bibtex-journal-key', 'journal_abbrev')
    >>> doc = from_data({'title': 'Hello', 'type': 'book', 'journal': 'jcp'})
    >>> import tempfile; doc.set_folder('path/to/superfolder')
    >>> to_bibtex(doc)
    '@book{superfolder,\\n  journal = { jcp },\\n  title = { Hello },\\n  type = { book },\\n}\\n'
    >>> doc['journal_abbrev'] = 'j'
    >>> to_bibtex(doc)
    '@book{superfolder,\\n  journal = { j },\\n  title = { Hello },\\n  type = { book },\\n}\\n'
    >>> del doc['title']
    >>> doc['ref'] = 'hello1992'
    >>> to_bibtex(doc)
    '@book{hello1992,\\n  journal = { j },\\n  type = { book },\\n}\\n'
    """
    bibtexString = ""
    bibtexType = ""

    # First the type, article ....
    if "type" in document.keys():
        if document["type"] in papis.bibtex.bibtex_types:
            bibtexType = document["type"]
        elif document["type"] in papis.bibtex.bibtex_type_converter.keys():
            bibtexType = papis.bibtex.bibtex_type_converter[document["type"]]
    if not bibtexType:
        bibtexType = "article"

    ref = document["ref"]
    if not ref:
        try:
            ref = os.path.basename(document.get_main_folder())
        except:
            if document.has('doi'):
                ref = document['doi']
            else:
                ref = 'noreference'

    ref = re.sub(r'[;,()\/{}\[\]]', '', ref)

    bibtexString += "@%s{%s,\n" % (bibtexType, ref)
    for bibKey in sorted(document.keys()):
        logger.debug('%s : %s' % (bibKey, document[bibKey]))
        if bibKey in papis.bibtex.bibtex_key_converter:
            newBibKey = papis.bibtex.bibtex_key_converter[bibKey]
            document[newBibKey] = document[bibKey]
            continue
        if bibKey in papis.bibtex.bibtex_keys:
            value = str(document[bibKey])
            if not papis.config.get('bibtex-unicode'):
                value = papis.bibtex.unicode_to_latex(value)
            if bibKey == 'journal':
                bibtex_journal_key = papis.config.get('bibtex-journal-key')
                if bibtex_journal_key in document.keys():
                    bibtexString += "  %s = { %s },\n" % (
                        'journal',
                        papis.bibtex.unicode_to_latex(
                            str(
                              document[papis.config.get('bibtex-journal-key')]
                            )
                        )
                    )
                elif bibtex_journal_key not in document.keys():
                    logger.warning(
                        "Key '%s' is not present for ref=%s" % (
                            papis.config.get('bibtex-journal-key'),
                            document["ref"]
                        )
                    )
                    bibtexString += "  %s = { %s },\n" % (
                        'journal',
                        value
                    )
            else:
                bibtexString += "  %s = { %s },\n" % (
                    bibKey,
                    value
                )
    bibtexString += "}\n"
    return bibtexString


def to_json(document):
    """Export information into a json string
    :param document: Papis document
    :type  document: Document
    :returns: Json formatted info file
    :rtype:  str

    >>> doc = from_data({'title': 'Hello World'})
    >>> to_json(doc)
    '{"title": "Hello World"}'
    """
    import json
    return json.dumps(to_dict(document))


def to_dict(document):
    """Gets a python dictionary with the information of the document
    :param document: Papis document
    :type  document: Document
    :returns: Python dictionary
    :rtype:  dict
    """
    result = dict()
    for key in document.keys():
        result[key] = document[key]
    return result


def dump(document):
    """Return information string without any obvious format
    :param document: Papis document
    :type  document: Document
    :returns: String with document's information
    :rtype:  str

    >>> doc = from_data({'title': 'Hello World'})
    >>> dump(doc)
    'title:   Hello World\\n'
    """
    string = ""
    for i in document.keys():
        string += str(i)+":   "+str(document[i])+"\n"
    return string


def delete(document):
    """This function deletes a document from disk.
    :param document: Papis document
    :type  document: papis.document.Document
    """
    import shutil
    folder = document.get_main_folder()
    shutil.rmtree(folder)


def move(document, path):
    """This function moves a document to path, it supposes that
    the document exists in the location ``document.get_main_folder()``.
    Warning: This method will change the folder in the document object too.
    :param document: Papis document
    :type  document: papis.document.Document
    :param path: Full path where the document will be moved to
    :type  path: str

    >>> doc = from_data({'title': 'Hello World'})
    >>> doc.set_folder('path/to/folder')
    >>> import tempfile; newfolder = tempfile.mkdtemp()
    >>> move(doc, newfolder)
    Traceback (most recent call last):
    ...
    Exception: There is already...
    """
    import shutil
    path = os.path.expanduser(path)
    if os.path.exists(path):
        raise Exception(
            "There is already a document in {0}, please check it,\n"
            "a temporary papis document has been stored in {1}".format(
                path, document.get_main_folder()
            )
        )
    shutil.move(document.get_main_folder(), path)
    # Let us chmod it because it might come from a temp folder
    # and temp folders are per default 0o600
    os.chmod(path, papis.config.getint('dir-umask'))
    document.set_folder(path)


class DocHtmlEscaped(dict):
    """
    Small helper class to escape html elements.

    >>> DocHtmlEscaped(from_data(dict(title='> >< int & "" "')))['title']
    '&gt; &gt;&lt; int &amp; &quot;&quot; &quot;'
    """

    def __init__(self, doc):
        self.doc = doc

    def __getitem__(self, key):
        return str(self.doc[key]).replace('&', '&amp;')\
                                .replace('<', '&lt;')\
                                .replace('>', '&gt;')\
                                .replace('"', '&quot;')


class Document(object):

    """Class implementing the entry abstraction of a document in a library.
    It is basically a python dictionary with more methods.
    """

    subfolder = ""
    _infoFilePath = ""

    def __init__(self, folder=None, data=None):
        self._keys = []
        self._folder = None

        if folder is not None:
            self.set_folder(folder)
            self.load()

        if data is not None:
            self.update(data)

    def __delitem__(self, key):
        """Deletes property from document, e.g. ``del doc['url']``.
        :param key: Name of the property.
        :type  key: str

        >>> doc = from_data({'title': 'Hello', 'type': 'book'})
        >>> del doc['title']
        >>> doc.has('title')
        False
        """
        self._keys.pop(self._keys.index(key))
        delattr(self, key)

    def __setitem__(self, key, value):
        """Sets property to value from document, e.g. ``doc['url'] =
        'www.gnu.org'``.
        :param key: Name of the property.
        :type  key: str
        :param value: Value of the parameter
        :type  value: str,int,float,list
        """
        self._keys.append(key)
        setattr(self, key, value)

    def __getitem__(self, key):
        """Gets property to value from document, e.g. ``a = doc['url']``.
        If the property `key` does not exist, then the empy string is returned.

        :param key: Name of the property.
        :type  key: str
        :returns: Value of the property
        :rtype:  str,int,float,list
        """
        return getattr(self, key) if hasattr(self, key) else ""

    @property
    def html_escape(self):
        return DocHtmlEscaped(self)

    def get_main_folder(self):
        """Get full path for the folder where the document and the information
        is stored.
        :returns: Folder path
        """
        return self._folder

    def set_folder(self, folder):
        """Set document's folder. The info_file path will be accordingly set.

        :param folder: Folder where the document will be stored, full path.
        :type  folder: str
        """
        self._folder = folder
        self._infoFilePath = os.path.join(
            folder,
            papis.utils.get_info_file_name()
        )
        self.subfolder = self.get_main_folder().replace(
            os.environ["HOME"], ""
        ).replace(
            "/", " "
        )

    def get_main_folder_name(self):
        """Get main folder name where the document and the information is
        stored.
        :returns: Folder name
        """
        return os.path.basename(self._folder)

    def has(self, key):
        """Check if the information file has some key defined.

        :param key: Key name to be checked
        :returns: True/False

        >>> doc = from_data({'title': 'Hello World'})
        >>> doc.has('title')
        True
        >>> doc.has('author')
        False
        """
        return key in self.keys()

    def rm_file(self, filepath):
        """Remove file from document, it also removes the entry in `files`

        :filepath: Full file path for file

        >>> doc = from_data({'title': 'Hello', 'files': ['a.pdf']})
        >>> doc.rm_file('b.pdf')
        Traceback (most recent call last):
        ...
        Exception: File b.pdf not tracked by document
        """
        basename = os.path.basename(filepath)
        if basename not in self['files']:
            raise Exception("File %s not tracked by document" % basename)
        os.remove(filepath)
        self['files'].pop(self['files'].index(basename))

    def rm(self):
        """Removes document's folder, effectively removing it from the library.

        >>> doc = from_data({'title': 'Hello World'})
        >>> import tempfile; doc.set_folder(tempfile.mkdtemp())
        >>> doc.rm()
        """
        shutil.rmtree(self.get_main_folder())

    def save(self):
        """Saves the current document's information into the info file.
        """
        import yaml
        fd = open(self._infoFilePath, "w+")
        structure = dict()
        for key in self.keys():
            structure[key] = self[key]
        # self.logger.debug("Saving %s " % self.get_info_file())
        yaml.dump(
            structure,
            fd,
            allow_unicode=papis.config.getboolean("info-allow-unicode"),
            default_flow_style=False
        )
        fd.close()

    def update(self, data, force=False, interactive=False):
        """Update document's information from an info dictionary.

        :param data: Dictionary with key and values to be updated
        :type  data: dict
        :param force: If True, the update turns into a replace, i.e., it
            replaces the old value by the new value stored in data.
        :type  force: bool
        :param interactive: If True, it will ask for user's input every time
            that the values differ.
        :type  interactive: bool

        """
        for key in data.keys():
            if self[key] != data[key]:
                if force:
                    self[key] = data[key]
                elif interactive:
                    confirmation = papis.utils.confirm(
                        "(%s conflict) Replace '%s' by '%s'?" % (
                            key, self[key], data[key]
                        )
                    )
                    if confirmation:
                        self[key] = data[key]
                elif self[key] is None or self[key] == '':
                    self[key] = data[key]

    def get_info_file(self):
        """Get full path for the info file
        :returns: Full path for the info file
        :rtype: str
        """
        return self._infoFilePath

    def get_files(self):
        """Get the files linked to the document, if any.

        :returns: List of full file paths
        :rtype:  list
        """
        files = self["files"] if isinstance(self["files"], list) \
            else [self["files"]]
        result = []
        for f in files:
            result.append(os.path.join(self.get_main_folder(), f))
        return result

    def keys(self):
        """Returns the keys defined for the document.

        :returns: Keys for the document
        :rtype:  list
        """
        return self._keys

    @property
    def has_citations(self):
        """Returns string defined in config if keys contains citations
        else returns None.

        :returns: String or None
        :rtype: str OR None

        >>> import papis.config
        >>> doc = from_data({'title': 'Hello World'})
        >>> doc.has_citations
        ''
        >>> doc.update(dict(citations=[]))
        >>> doc.has_citations == papis.config.get('citation-string')
        True
        """

        if 'citations' in self.keys():
            return papis.config.get('citation-string')
        else:
            return ''

    def dump(self):
        """Return information string without any obvious format
        :returns: String with document's information
        :rtype:  str

        >>> doc = from_data({'title': 'Hello World'})
        >>> doc.dump()
        'title:   Hello World\\n'
        """
        return dump(self)

    def load(self):
        """Load information from info file
        """
        import yaml
        # TODO: think about if it's better to raise an exception here
        # TODO: if no info file is found
        try:
            fd = open(self.get_info_file(), "r")
        except:
            return False
        try:
            structure = yaml.load(fd)
            for key in structure:
                self[key] = structure[key]
            fd.close()
        except Exception as e:
            logging.error(
                'Error reading yaml file in {0}'.format(self.get_info_file()) +
                '\nPlease check it!\n\n{0}'.format(str(e))
            )
            fd.close()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import logging
import papis.commands.default


logger = logging.getLogger("papis")
if "--debug" in sys.argv:
    log_format = '%(relativeCreated)d-'
    log_format += '%(levelname)s:%(name)s:%(message)s'
    logging.basicConfig(level=logging.DEBUG, format=log_format)
    sys.argv.pop(sys.argv.index("--debug"))
    logger.debug("DEBUG MODE FOR DEVELOPERS ON")


logger.debug("Imported commands")


def main():
    papis.commands.default.run()



"""
This command is to list contents of a library.

CLI Examples
^^^^^^^^^^^^

- List all document files associated will all entries:

    .. code:: bash

        papis list --file

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/XwD0ZaUORoOonwDw4rXoQDkjZ.js"
        id="asciicast-XwD0ZaUORoOonwDw4rXoQDkjZ" async></script>

- List all document year and title with custom formatting:

    .. code:: bash

        papis list --format '{doc[year]} {doc[title]}'

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/NZ8Ii1wWYPo477CIL4vZhUqOy.js"
        id="asciicast-NZ8Ii1wWYPo477CIL4vZhUqOy" async></script>

- List all documents according to the bibitem formatting (stored in a template
  file ``bibitem.template``):

    .. code:: bash

        papis list --template bibitem.template

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/QZTBZ3tFfyk9WQuJ9WWB2UpSw.js"
        id="asciicast-QZTBZ3tFfyk9WQuJ9WWB2UpSw" async></script>

Python examples
^^^^^^^^^^^^^^^

.. code:: python

    # Import the run function from the list command

    from papis.commands.list import run as papis_list

    documents = papis_list(query='einstein', library='papis')

    for doc in documents:
        print(doc["url"])

    # etc...
    info_files = list(query='einstein', library='papis', info_files=True)

    # do something with the info file paths...

Cli
^^^
.. click:: papis.commands.list:cli
    :prog: papis list
"""

import logging
import papis
from papis.api import status
import os
import papis.utils
import papis.config
import papis.database
import papis.downloaders.utils
import papis.cli
import click

logger = logging.getLogger('list')


def run(
        query="",
        library=None,
        libraries=False,
        downloaders=False,
        pick=False,
        files=False,
        folders=False,
        info_files=False,
        fmt="",
        template=None
        ):
    """Main method to the list command

    :returns: List different objects
    :rtype:  list
    """
    if library is None:
        library = papis.config.get_lib()
    config = papis.config.get_configuration()
    db = papis.database.get(library)
    if template is not None:
        if not os.path.exists(template):
            logger.error(
                "Template file %s not found" % template
            )
            return status.file_not_found
        fd = open(template)
        fmt = fd.read()
        fd.close()

    if downloaders:
        return papis.downloaders.utils.get_available_downloaders()

    if libraries:
        return [
            config[section]['dir']
            for section in config
            if 'dir' in config[section]
        ]

    documents = db.query(query)

    if pick:
        documents = filter(lambda x: x, [papis.api.pick_doc(documents)])

    if files:
        return [
            doc_file for files in [
                document.get_files() for document in documents
            ] for doc_file in files
        ]
    elif info_files:
        return [
            os.path.join(
                document.get_main_folder(),
                document.get_info_file()
            ) for document in documents
        ]
    elif fmt:
        return [
            papis.utils.format_doc(fmt, document)
            for document in documents
        ]
    elif folders:
        return [
            document.get_main_folder() for document in documents
        ]
    else:
        return documents


@click.command()
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    "-i",
    "--info",
    help="Show the info file name associated with the document",
    default=False,
    is_flag=True
)
@click.option(
    "-f",
    "--file",
    help="Show the file name associated with the document",
    default=False,
    is_flag=True
)
@click.option(
    "-d",
    "--dir",
    help="Show the folder name associated with the document",
    default=False,
    is_flag=True
)
@click.option(
    "--format",
    help="List entries using a custom papis format, e.g."
    " '{doc[year] {doc[title]}",
    default=None
)
@click.option(
    "--template",
    help="Template file containing a papis format to list entries",
    default=None
)
@click.option(
    "-p",
    "--pick",
    help="Pick the document instead of listing everything",
    default=False,
    is_flag=True
)
@click.option(
    "--downloaders",
    help="List available downloaders",
    default=False,
    is_flag=True
)
@click.option(
    "--libraries",
    help="List defined libraries",
    default=False,
    is_flag=True
)
def cli(
        query,
        info,
        file,
        dir,
        format,
        template,
        pick,
        downloaders,
        libraries
        ):
    """List documents' properties"""

    if not libraries and not downloaders and not file and not info and not dir:
        dir = True

    lib = papis.config.get_lib()

    objects = run(
        query=query,
        library=lib,
        libraries=libraries,
        downloaders=downloaders,
        pick=pick,
        files=file,
        folders=dir,
        info_files=info,
        fmt=format,
        template=template
    )
    for o in objects:
        click.echo(o)
    return status.success



"""
The add command is one of the central commands of the papis command line
interface. It is a very versatile command with a fair amount of options.

There are also customization settings availabe for this command, check out
the :ref:`configuration page <add-command-options>` for add.

Examples
^^^^^^^^

- Add a document located in ``~/Documents/interesting.pdf`` and name the
  folder where it will be stored in the database ``interesting-paper-2021``

    .. code::

        papis add ~/Documents/interesting.pdf --name interesting-paper-2021

  if you want to add directly some key values, like ``author``, ``title``
  and ``tags``, you can also run the following:

    .. code::

        papis add ~/Documents/interesting.pdf \\
            --name interesting-paper-2021 \\
            --set author 'John Smith' \\
            --set title 'The interesting life of bees' \\
            --set year 1985 \\
            --set tags 'biology interesting bees'

- Add a paper that you have locally in a file and get the paper information
  through its ``doi`` identifier (in this case ``10.10763/1.3237134`` as an
  example):

    .. code::

        papis add ~/Documents/interesting.pdf --from-doi 10.10763/1.3237134

- Add paper to a library named ``machine-learning`` from ``arxiv.org``

    .. code::

        papis -l machine-learning add \\
            --from-url https://arxiv.org/abs/1712.03134

- If you do not want copy the original pdfs into the library, you can
  also tell papis to just create a link to them, for example

    .. code::

        papis add --link ~/Documents/interesting.pdf \\
            --from-doi 10.10763/1.3237134

  will add an entry into the papis library, but the pdf document will remain
  at ``~/Documents/interesting.pdf``, and in the document's folder
  there will be a link to ``~/Documents/interesting.pdf`` instead of the
  file itself. Of course you always have to be sure that the
  document at ``~/Documents/interesting.pdf`` does not disappear, otherwise
  you will end up without a document to open.


Examples in python
^^^^^^^^^^^^^^^^^^

There is a python function in the add module that can be used to interact
in a more effective way in python scripts,

.. autofunction:: papis.commands.add.run

Cli
^^^
.. click:: papis.commands.add:cli
    :prog: papis add

"""
import logging
import papis
from string import ascii_lowercase
import os
import re
import tempfile
import hashlib
import shutil
import subprocess
import papis.api
from papis.api import status
import papis.utils
import papis.config
import papis.bibtex
import papis.document
import papis.downloaders.utils
import papis.cli
import click
import builtins

logger = logging.getLogger('add')


def get_file_name(data, original_filepath, suffix=""):
    """Generates file name for the document

    :param data: Data parsed for the actual document
    :type  data: dict
    :param original_filepath: The full path to the original file
    :type  original_filepath: str
    :param suffix: Possible suffix to be appended to the file without
        its extension.
    :type  suffix: str
    :returns: New file name
    :rtype:  str

    """

    basename_limit = 150
    file_name_opt = papis.config.get('file-name')

    if file_name_opt is None:
        file_name_opt = os.path.basename(original_filepath)

    file_name_base = papis.utils.format_doc(
        file_name_opt,
        papis.document.from_data(data)
    )

    if len(file_name_base) > basename_limit:
        logger.warning(
            'Shortening the documents\' {} name for portability'.format(
                original_filepath
            )
        )
        file_name_base = file_name_base[0:basename_limit]

    filename_basename = papis.utils.clean_document_name(
        "{}{}".format(
            file_name_base,
            "-" + suffix if len(suffix) > 0 else ""
        )
    )

    filename = "{}.{}".format(
        filename_basename,
        papis.utils.get_document_extension(original_filepath)
    )

    return filename


def get_hash_folder(data, document_paths):
    """Folder name where the document will be stored.

    :data: Data parsed for the actual document
    :document_paths: Path of the document

    """
    import random
    author = "-{:.20}".format(data["author"])\
             if "author" in data.keys() else ""

    document_strings = b''
    for docpath in document_paths:
        with open(docpath, 'rb') as fd:
            document_strings += fd.read(2000)

    md5 = hashlib.md5(
        ''.join(document_paths).encode() +
        str(data).encode() +
        str(random.random()).encode() +
        document_strings
    ).hexdigest()

    result = md5 + author
    result = papis.utils.clean_document_name(result)

    return result


def get_document_extension(document_path):
    """Get document extension

    :document_path: Path of the document
    :returns: Extension (string)

    """
    import filetype
    filetype.guess(document_path)
    kind = filetype.guess(document_path)
    if kind is None:
        m = re.match(r"^.*\.([^.]+)$", os.path.basename(document_path))
        return m.group(1) if m else 'data'
    else:
        return kind.extension


def get_default_title(data, document_path, interactive=False):
    """
    >>> get_default_title({'title': 'hello world'}, 'whatever.pdf')
    'hello world'
    >>> open('Luces-de-bohemia.pdf', 'w+').close()
    >>> get_default_title(dict(), 'Luces-de-bohemia.pdf')
    'Luces de bohemia'
    """
    if "title" in data.keys():
        return data["title"]
    extension = get_document_extension(document_path)
    title = os.path.basename(document_path)\
        .replace("."+extension, "")\
        .replace("_", " ")\
        .replace("-", " ")
    if interactive:
        title = papis.utils.input(
            'Title?', title
        )
    return title


def get_default_author(data, document_path, interactive=False):
    """
    >>> get_default_author({'author': 'Garcilaso de la vega'}, 'whatever.pdf')
    'Garcilaso de la vega'
    >>> get_default_author(dict(), 'Luces-de-bohemia.pdf')
    'Unknown'
    """
    if "author" in data.keys():
        return data["author"]
    author = "Unknown"
    if interactive:
        author = papis.utils.input(
            'Author?', author
        )
    return author


def run(
        paths,
        data=dict(),
        name=None,
        file_name=None,
        subfolder=None,
        interactive=False,
        confirm=False,
        open_file=False,
        edit=False,
        commit=False,
        link=False
        ):
    """
    :param paths: Paths to the documents to be added
    :type  paths: []
    :param data: Data for the document to be added.
        If more data is to be retrieved from other sources, the data dictionary
        will be updated from these sources.
    :type  data: dict
    :param name: Name of the folder where the document will be stored
    :type  name: str
    :param file_name: File name of the document's files to be stored.
    :type  file_name: str
    :param subfolder: Folder within the library where the document's folder
        should be stored.
    :type  subfolder: str
    :param interactive: Wether or not interactive functionality of this command
        should be activated.
    :type  interactive: bool
    :param confirm: Wether or not to ask user for confirmation before adding.
    :type  confirm: bool
    :param open_file: Wether or not to ask user for opening file before adding.
    :type  open_file: bool
    :param edit: Wether or not to ask user for editing the infor file
        before adding.
    :type  edit: bool
    :param commit: Wether or not to ask user for committing before adding,
        in the case of course that the library is a git repository.
    :type  commit: bool
    """

    logger = logging.getLogger('add:run')
    # The folder name of the new document that will be created
    out_folder_name = None
    # The real paths of the documents to be added
    in_documents_paths = paths
    # The basenames of the documents to be added
    in_documents_names = []
    # The folder name of the temporary document to be created
    temp_dir = tempfile.mkdtemp()

    for p in in_documents_paths:
        if not os.path.exists(p):
            raise IOError('Document %s not found' % p)

    in_documents_names = [
        papis.utils.clean_document_name(doc_path)
        for doc_path in in_documents_paths
    ]

    tmp_document = papis.document.Document(temp_dir)

    if not name:
        logger.info("Getting an automatic name")
        out_folder_name = get_hash_folder(data, in_documents_paths)
    else:
        temp_doc = papis.document.Document(data=data)
        out_folder_name = papis.utils.format_doc(name, temp_doc)
        out_folder_name = papis.utils.clean_document_name(out_folder_name)
        del temp_doc

    data["files"] = in_documents_names
    out_folder_path = os.path.expanduser(os.path.join(
        papis.config.get('dir'), subfolder or '',  out_folder_name
    ))

    logger.debug("Folder name = % s" % out_folder_name)
    logger.debug("Folder path = % s" % out_folder_path)
    logger.debug("File(s)     = % s" % in_documents_paths)

    # First prepare everything in the temporary directory
    g = papis.utils.create_identifier(ascii_lowercase)
    string_append = ''
    if file_name is not None:  # Use args if set
        papis.config.set("file-name", file_name)
    new_file_list = []

    for in_file_path in in_documents_paths:

        # Rename the file in the staging area
        new_filename = papis.utils.clean_document_name(
            get_file_name(
                data,
                in_file_path,
                suffix=string_append
            )
        )
        new_file_list.append(new_filename)

        tmp_end_filepath = os.path.join(
            tmp_document.get_main_folder(),
            new_filename
        )
        string_append = next(g)

        if link:
            in_file_abspath = os.path.abspath(in_file_path)
            logger.debug(
                "[SYMLINK] '%s' to '%s'" %
                (in_file_abspath, tmp_end_filepath)
            )
            os.symlink(in_file_abspath, tmp_end_filepath)
        else:
            logger.debug(
                "[CP] '%s' to '%s'" %
                (in_file_path, tmp_end_filepath)
            )
            shutil.copy(in_file_path, tmp_end_filepath)

    data['files'] = new_file_list

    # Check if the user wants to edit before submitting the doc
    # to the library
    if edit:
        tmp_document.update(data, force=True)
        tmp_document.save()
        logger.info("Editing file before adding it")
        papis.api.edit_file(tmp_document.get_info_file(), wait=True)
        logger.info("Loading the changes made by editing")
        tmp_document.load()
        data = papis.document.to_dict(tmp_document)

    # Duplication checking
    logger.info("Check if the added document is already existing")
    try:
        found_document = papis.utils.locate_document_in_lib(tmp_document)
    except IndexError:
        logger.info("Document not found in library")
    else:
        logger.warning(
            '\n\n\n\n{0}\n\n\n\n'.format(papis.document.dump(found_document))
        )
        logger.warning("DUPLICATION WARNING")
        logger.warning(
            "The document above is in your library and it seems to match"
        )
        logger.warning(
            "the one that you're trying to add, "
            "I will prompt you for confirmation"
        )
        logger.warning(
            "(Hint) Use the update command if you just want to update"
            " the info."
        )
        confirm = True

    tmp_document.update(data, force=True)
    tmp_document.save()
    if open_file:
        for d_path in tmp_document.get_files():
            papis.api.open_file(d_path)
    if confirm:
        if not papis.utils.confirm('Really add?'):
            return status.success

    logger.info(
        "[MV] '%s' to '%s'" %
        (tmp_document.get_main_folder(), out_folder_path)
    )
    # This also sets the folder of tmp_document
    papis.document.move(tmp_document, out_folder_path)
    papis.database.get().add(tmp_document)
    if commit:
        subprocess.call(["git", "-C", out_folder_path, "add", "."])
        subprocess.call(
            ["git", "-C", out_folder_path, "commit", "-m", "Add document"]
        )
    return status.success


@click.command(
    help="Add a document into a given library"
)
@click.help_option('--help', '-h')
@click.argument("files", type=click.Path(exists=True), nargs=-1)
@click.option(
    "-s", "--set", "set_list",
    help="Set some information before",
    multiple=True,
    type=(str, str)
)
@click.option(
    "-d", "--dir", "directory",
    help="Subfolder in the library",
    default=""
)
@click.option(
    "-i", "--interactive/--no-interactive",
    help="Do some of the actions interactively",
    default=lambda: True if papis.config.get('add-interactive') else False
)
@click.option(
    "--name",
    help="Name for the document's folder (papis format)",
    default=lambda: papis.config.get('add-name')
)
@click.option(
    "--file-name",
    help="File name for the document (papis format)",
    default=None
)
@click.option(
    "--from-bibtex",
    help="Parse information from a bibtex file",
    default=""
)
@click.option(
    "--from-yaml",
    help="Parse information from a yaml file",
    default=""
)
@click.option(
    "--from-folder",
    help="Add document from folder being a valid papis document"
         " (containing info.yaml)",
    default=""
)
@click.option(
    "--from-url",
    help="Get document and information from a"
    "given url, a parser must be implemented",
    default=""
)
@click.option(
    "--from-doi",
    help="Doi to try to get information from",
    default=None
)
@click.option(
    "--from-pmid",
    help="PMID to try to get information from",
    default=None
)
@click.option(
    "--from-lib",
    help="Add document from another library",
    default=""
)
@click.option(
    "--confirm/--no-confirm",
    help="Ask to confirm before adding to the collection",
    default=lambda: True if papis.config.get('add-confirm') else False
)
@click.option(
    "--open/--no-open", "open_file",
    help="Open file before adding document",
    default=lambda: True if papis.config.get('add-open') else False
)
@click.option(
    "--edit/--no-edit",
    help="Edit info file before adding document",
    default=lambda: True if papis.config.get('add-edit') else False
)
@click.option(
    "--commit/--no-commit",
    help="Commit document if library is a git repository",
    default=False
)
@click.option(
    "--link/--no-link",
    help="Instead of copying the file to the library, create a link to"
         "its original location",
    default=False
)
@click.option(
    #TODO: REMOVE IT AT SOME POINT
    "--no-document",
    default=False,
    is_flag=True,
    help="DEPRECATION NOTICE: This option is no longer valid, "
         "it will be removed in future releases"
)
def cli(
        files,
        set_list,
        directory,
        interactive,
        name,
        file_name,
        from_bibtex,
        from_yaml,
        from_folder,
        from_url,
        from_doi,
        from_pmid,
        from_lib,
        confirm,
        open_file,
        edit,
        commit,
        link,
        no_document
        ):
    """
    :param from_folder: Filepath where to find a papis document (folder +
        info file) to be added to the library.
    :type  from_folder: str
    :param from_doi: doi number to try to download information from.
    :type  from_doi: str
    :param from_pmid: pmid number to try to download information from.
    :type  from_pmid: str
    :param from_url: Url to try to download information and files from.
    :type  from_url: str
    :param from_bibtex: Filepath where to find a file containing bibtex info.
    :type  from_bibtex: str
    :param from_yaml: Filepath where to find a file containing yaml info.
    :type  from_yaml: str
    """
    data = dict()
    files = list(files)

    for data_set in set_list:
        data[data_set[0]] = data_set[1]

    logger = logging.getLogger('cli:add')

    if from_lib:
        doc = papis.api.pick_doc(
            papis.api.get_all_documents_in_lib(from_lib)
        )
        if doc:
            from_folder = doc.get_main_folder()

    try:
        # Try getting title if title is an argument of add
        data["title"] = data.get('title') or get_default_title(
            data,
            files[0],
            interactive
        )
        logger.debug("Title = % s" % data["title"])
    except:
        pass

    try:
        # Try getting author if author is an argument of add
        data["author"] = data.get('author') or get_default_author(
            data,
            files[0],
            interactive
        )
        logger.debug("Author = % s" % data["author"])
    except:
        pass

    if from_folder:
        original_document = papis.document.Document(from_folder)
        from_yaml = original_document.get_info_file()
        files.extend(original_document.get_files())

    if from_url:
        logger.info("Attempting to retrieve from url")
        url_data = papis.downloaders.utils.get(from_url)
        data.update(url_data["data"])
        files.extend(url_data["documents_paths"])
        # If no data was retrieved and doi was found, try to get
        # information with the document's doi
        if not data and url_data["doi"] is not None and\
                not from_doi:
            logger.warning(
                "I could not get any data from %s" % from_url
            )
            from_doi = url_data["doi"]

    if from_pmid:
        logger.info("Using PMID %s via HubMed" % from_pmid)
        hubmed_url = "http://pubmed.macropus.org/articles/"\
                     "?format=text%%2Fbibtex&id=%s" % from_pmid
        bibtex_data = papis.downloaders.utils.get_downloader(
            hubmed_url,
            "get"
        ).get_document_data().decode("utf-8")
        bibtex_data = papis.bibtex.bibtex_to_dict(bibtex_data)
        if len(bibtex_data):
            data.update(bibtex_data[0])
            if "doi" in data and not from_doi:
                from_doi = data["doi"]
        else:
            logger.error("PMID %s not found or invalid" % from_pmid)

    if from_doi:
        logger.info("I'll try using doi %s" % from_doi)
        data.update(papis.utils.doi_to_data(from_doi))
        if len(files) == 0 and \
                papis.config.get('doc-url-key-name') in data.keys():

            doc_url = data[papis.config.get('doc-url-key-name')]
            logger.info(
                'You did not provide any files, but I found a possible '
                'url where the file might be'
            )
            logger.info(
                'I am trying to download the document from %s' % doc_url
            )
            down = papis.downloaders.utils.get_downloader(doc_url, 'get')
            assert(down is not None)
            tmp_filepath = tempfile.mktemp()
            logger.debug("Saving in %s" % tmp_filepath)

            with builtins.open(tmp_filepath, 'wb+') as fd:
                fd.write(down.get_document_data())

            logger.info('Opening the file')
            papis.api.open_file(tmp_filepath)
            if papis.utils.confirm('Do you want to use this file?'):
                files.append(tmp_filepath)

    if from_yaml:
        logger.info("Reading yaml input file = %s" % from_yaml)
        data.update(papis.utils.yaml_to_data(from_yaml))

    if from_bibtex:
        logger.info("Reading bibtex input file = %s" % from_bibtex)
        bib_data = papis.bibtex.bibtex_to_dict(from_bibtex)
        if len(bib_data) > 1:
            logger.warning(
                'Your bibtex file contains more than one entry,'
                ' I will be taking the first entry'
            )
        data.update(bib_data[0])

    assert(isinstance(data, dict))

    return run(
        list(files),
        data=data,
        name=name,
        file_name=file_name,
        subfolder=directory,
        interactive=interactive,
        confirm=confirm,
        open_file=open_file,
        edit=edit,
        commit=commit,
        link=link
    )



import logging
logger = logging.getLogger('database')

DATABASES = dict()


def get(library=None):
    global DATABASES
    import papis.config
    backend = papis.config.get('database-backend')
    if library is None:
        library = papis.config.get_lib()
    database = DATABASES.get(library)
    # if there is already a database and the backend of the database
    # is the same as the config backend, then return that library
    # else we will (re)define the database in the dictionary DATABASES
    if database is not None and database.get_backend_name() == backend:
        return DATABASES.get(library)
    if backend == "papis":
        import papis.database.cache
        DATABASES[library] = papis.database.cache.Database(library)
        return DATABASES.get(library)
    elif backend == "whoosh":
        import papis.database.whoosh
        DATABASES[library] = papis.database.whoosh.Database(library)
        return DATABASES.get(library)
    else:
        raise Exception('No valid database type: {}'.format(backend))


def get_all_query_string():
    return get().get_all_query_string()


def clear_cached():
    global DATABASES
    DATABASES = dict()



"""This command edits the information of the documents.
The editor used is defined by the ``editor`` configuration setting.


Cli
^^^
.. click:: papis.commands.edit:cli
    :prog: papis edit
"""
import papis
import os
import papis.api
import papis.utils
import papis.config
import papis.database
import papis.cli
import click
import logging


def run(document, wait=True):
    database = papis.database.get()
    papis.utils.general_open(document.get_info_file(), "editor", wait=wait)
    document.load()
    database.update(document)


@click.command()
@click.help_option('-h', '--help')
@papis.cli.query_option()
@click.option(
    "-n",
    "--notes",
    help="Edit notes associated to the document",
    default=False,
    is_flag=True
)
@click.option(
    "--all",
    help="Edit all matching documents",
    default=False,
    is_flag=True
)
@click.option(
    "-e",
    "--editor",
    help="Editor to be used",
    default=None
)
def cli(
        query,
        notes,
        all,
        editor
        ):
    """Edit document information from a given library"""

    logger = logging.getLogger('cli:edit')
    documents = papis.database.get().query(query)

    if editor is not None:
        papis.config.set('editor', editor)

    if not all:
        document = papis.api.pick_doc(documents)
        documents = [document] if document else []

    if len(documents) == 0:
        return 0

    for document in documents:
        if notes:
            logger.debug("Editing notes")
            if not document.has("notes"):
                logger.warning(
                    "The document selected has no notes attached, \n"
                    "creating a notes files"
                )
                document["notes"] = papis.config.get("notes-name")
                document.save()
            notesPath = os.path.join(
                document.get_main_folder(),
                document["notes"]
            )

            if not os.path.exists(notesPath):
                logger.debug("Creating %s" % notesPath)
                open(notesPath, "w+").close()

            papis.api.edit_file(notesPath)
        else:
            run(document)



"""General
*******

.. papis-config:: local-config-file

    Name AND relative path of the local configuration file that papis
    will additionally read if the file is present in the current
    directory or in the base directory of a given library.

    This is useful for instance if you have a library somewhere
    for which you want special configuration settings to be set
    but you do not want these settings to cluster in your configuration
    file. It is also useful if you're sharing a library with someone
    else and you want them to have the same settings in that library as
    you. Imagine you're sharing a library of datasheets with your friend
    Fulano. You have your library at

    ::

        ~/Documents/lib-with-fulano

    and you've set a local configuration file there

    ::

        ~/Documents/lib-with-fulano/.papis.config

    then whenever Fulano uses that library and the file is also present,
    his papis program will also read the configuration settings at
    the path above.

.. papis-config:: dir-umask

    This is the default ``umask`` that will be used to create the new
    documents' directories.

.. papis-config:: use-git

    Some commands will issue git commands if this option is set to ``True``.
    For example in ``mv`` or ``rename``.

.. papis-config:: browse-query-format

    The query string that is to be searched for in the ``browse`` command
    whenever a search engine is used.

.. papis-config:: search-engine

    Search engine to be used by some commands like ``browse``.

.. papis-config:: user-agent

    User agent used by papis whenever it obtains information from external
    servers.

.. papis-config:: scripts-short-help-regex

    This is the format of the short help indicator in external papis
    commands.

.. papis-config:: info-name

    The default name of the information files.

.. papis-config:: doc-url-key-name

    Some documents might have associated apart from an url also a file url,
    the key name appearing in the information file is defined by
    this setting.

.. papis-config:: default-library

    The name of the library that is to be searched when ``papis``
    is run without library arguments.

.. papis-config:: export-text-format

    The default output papis format for ``papis export --text``.

.. papis-config:: format-doc-name

    This setting controls the name of the document in the papis format strings
    like in format strings such as ``match-format`` or ``header-format``.
    For instance, if you are managing videos, you might want to
    set this option to ``vid`` in order to set  the ``header-format`` to
    ``{doc[title]} - {doc[director]} - {doc[duration]}``.

.. papis-config:: match-format

    Default format that is used to match a document against in order to select
    it. For example if the ``match-format`` is equal to
    ``{doc[year]} {doc[author]}`` then title of a document will not work
    to match a document, only the year and author.

.. papis-config:: header-format

    Default format that is used to show a document in order to select it.

.. papis-config:: format-jinja2-enable

    This setting is to enable the `jinja2 <http://jinja.pocoo.org//>`_ template
    engine to render the papis templates being used, as ``header-format``,
    ``match-format`` etc...

    For instance you could set the option ``header-format`` to

    .. code:: html

        <span color='#ff00ff'>{{doc.html_escape["title"]}}</span>
        <span color='#00ff00'>  {{doc.html_escape["author"]}}</span>
        <span color='#00ffaa'>   ({{doc.html_escape["year"]}}) </span>
        {%- if doc.has('tags') %}<span>[<yellow>{{doc['tags']}}</yellow>] </span>{%- endif %}
        {%- if doc.has('citations') %}<red>{{doc['citations']|length}}</red>{%- endif %}
        {%- if doc.has('url') %}
        <span>    {{doc.html_escape["url"]}}</span>
        {%- endif %}

    To use it, just install jinja2.

.. papis-config:: header-format-file

    This option should have the path of a file with the ``header-format``
    template. Sometimes templates can get big so this is a way
    of not cluttering the config file with text.

    As an example you would set

    .. code:: ini

        [papers]

        header-format-file = ~/.papis/config/styles/header.txt

.. papis-config:: info-allow-unicode

    This flag is to be set if you want to allow unicode characters
    in your info file or not. If it is set to false then a representation
    for the unicode characters will be written in its place.
    Since we should be living in an unicode world, it is set to ``True``
    by default.

Tools options
^^^^^^^^^^^^^

.. papis-config:: opentool

    This is the general program that will be used to open documents.
    As for now papis is not intended to detect the type of document to be open
    and decide upon it how to open the document. You should set this
    to the right program for the tool. If you are in linux you might want
    to take a look at `ranger <http://ranger.nongnu.org>`_ or let
    the default handle it in your system.
    For mac users you might set this to ``open``.

.. papis-config:: browser
    :default: $BROWSER

    Program to be used for opening websites, the default is the environment
    variable ``$BROWSER``.

.. papis-config:: picktool

    This is the program used whenever papis asks you to pick a document
    or options in general.

    Possible options are:
        - papis.pick
        - vim
        - dmenu

.. papis-config:: editor
    :default: $EDITOR

    Editor used to edit files in papis, for instance for the ``papis edit``
    command. It defaults to the ``$EDITOR`` environment variable, if this is
    not set then it will default to the ``$VISUAL`` environment variable.
    Otherwise the default editor in your system will be used.

.. papis-config:: xeditor

    Sometimes papis might use an editor that uses a windowing system
    (GUI Editor), you can set this to your preferred gui based editor, e.g.
    ``gedit``, ``xemacs``, ``gvim`` to name a few.


.. papis-config:: file-browser

    File browser to be used when opening a directory, it defaults to the
    default file browser in your system, however you can set it to different
    file browsers such as ``dolphin``, ``thunar``, ``ranger`` to name a few.


Bibtex options
^^^^^^^^^^^^^^

.. papis-config:: bibtex-journal-key

    Journal publishers may request abbreviated journal titles. This
    option allows the user to set the key for the journal entry when using
    ``papis export --bibtex``.

    Set as ``full_journal_title`` or ``abbrev_journal_title`` for
    whichever style required. Default is ``journal``.

.. papis-config:: extra-bibtex-keys

    When exporting documents in bibtex format, you might want to add
    non-standard bibtex keys such as ``doc_url`` or ``tags``, you can add
    these here as comma separated values, for example
    ``extra-bibtex-keys = tags, doc_url``.

.. papis-config:: extra-bibtex-types

    Allow non-standard bibtex types to be recognized, e.g,
    ``extra-bibtex-types = wikipedia, video, song``.
    See `bibtex reference
        <http://mirror.easyname.at/ctan/biblio/bibtex/base/btxdoc.pdf>`_.

.. papis-config:: multiple-authors-format

    When retrieving automatic author information from services like
    crossref.org, papis usually builds the ``author`` field for the
    given document. The format how every single author name is built
    is given by this setting, for instance you could customize it
    by the following:

    ::

        multiple-authors-format = {au[surname]} -- {au[given_name]}

    which would given in the case of Albert Einstein the string
    ``Einstein -- Albert``.

.. papis-config:: multiple-authors-separator

    Similarly to ``multiple-authors-format``, this is the string that
    separates single authors in the ``author`` field. If it is set to
    `` and `` then you would have ``<author 1> and <author 2> and ....``
    in the ``author`` field.

.. papis-config:: bibtex-unicode

    Wether or not to allow direct unicode characters in the document
    fields to be exported into the bibtex text.

.. _add-command-options:

Add command options
^^^^^^^^^^^^^^^^^^^

.. papis-config:: ref-format

    This flag is set to change the ``ref`` flag in the info.yaml file
    when a document is imported. For example: I prefer the format
    FirstAuthorYear e.g. Plews2019. This would be achieved by the
    following:

    ::

        ref-format = {doc[author_list][0][surname]}{doc[year]}

    The default behavior is to set the doi as the ref.

.. papis-config:: add-confirm

    If set to ``True``, everytime you run ``papis add``
    the flag ``--confirm`` will be added automatically. If is set to ``True``
    and you add it, i.e., you run ``papis add --confirm``, then it will
    fave the contrary effect, i.e., it will not ask for confirmation.

.. papis-config:: add-name
    :default: empty string

    Default name for the folder of newly added documents. For example, if you
    the folder of your documents to be named after the format
    ``author-title`` then you should set it to
    the papis format: ``{doc[author]}-{doc[title]}``.
    Per default a hash followed by the author name is created.

.. papis-config:: file-name

    Same as ``add-name``, but for files, not folders. If it is not set,
    the names of the files will be cleaned and taken `as-is`.

.. papis-config:: add-interactive

    If set to ``True``, everytime you run ``papis add``
    the flag ``--interactive`` will be added automatically. If is set to
    ``True`` and you add it, i.e., you run ``papis add --interactive``, then it
    will fave the contrary effect, i.e., it will not run in interactive mode.

.. papis-config:: add-edit

    If set to ``True``, everytime you run ``papis add``
    the flag ``--edit`` will be added automatically. If is set to
    ``True`` and you add it, i.e., you run ``papis add --edit``, then it
    will fave the contrary effect, i.e., it will not prompt to edit the info
    file.

.. papis-config:: add-open

    If set to ``True``, everytime you run ``papis add``
    the flag ``--open`` will be added automatically. If is set to
    ``True`` and you add it, i.e., you run ``papis add --open``, then it
    will fave the contrary effect, i.e., it will not open the attached files
    before adding the document to the library.

Browse command options
^^^^^^^^^^^^^^^^^^^^^^

.. papis-config:: browse-key

    This command provides the key that is used to generate the
    url. For users that ``papis add --from-doi``, setting browse-key
    to ``doi`` constructs the url from dx.doi.org/DOI, providing a
    much more accurate url.

    Default value is set to ``url``. If the user needs functionality
    with the ``search-engine`` option, set the option to an empty
    string e.g.  ::

        browse-key = ''

.. _edit-command-options:

Edit command options
^^^^^^^^^^^^^^^^^^^^

.. papis-config:: notes-name

    In ``papis edit`` you can edit notes about the document. ``notes-name``
    is the default name of the notes file, which by default is supposed
    to be a TeX file.

.. _marks-options:

Marks
^^^^^

.. papis-config:: open-mark

    If this option is set to ``True``, then every time that papis opens
    a document it will ask to open a mark first.
    If it is set to ``False``, then doing

    .. code::

        papis open --mark

    will avoid opening a mark.

.. papis-config:: mark-key-name

    This is the default key name for the marks in the info file, for
    example if you set ``mark-key-name = bookmarks`` then you would have
    in your ``info.yaml`` file

    .. code::

        author: J. Krishnamurti
        bookmarks:
        - name: Chapter 1
          value: 120

.. papis-config:: mark-format-name

    This is the name of the mark to be passed to the options
    ``mark-header-format`` etc... E.g. if you set ``mark-format-name = m``
    then you could set ``mark-header-format = {m[value]} - {m[name]}``.

.. papis-config:: mark-header-format

    This is the format in which the mark will appear whenever the user
    has to pick one, you can change this in order to make ``marks`` work
    in the way you like. Per default it is assumed that every mark
    has a ``name`` and a ``value`` key, but this you can change.

.. papis-config:: mark-match-format

    Format in which the mark name has to match the user input.

.. papis-config:: mark-opener-format

    Due to the difficulty to generalize opening a general document
    at a given bookmark, the user should set this in whichever way
    it suits their needs. For example

    - If you are using the pdf viewer ``evince`` and you want to open a
      mark, you would use

        ::

            mark-opener-format = evince -p {mark[value]}

    - If you are using ``okular`` you would use

        ::

            mark-opener-format = okular -p {mark[value]}

    - If you are using ``zathura``, do

        ::

            mark-opener-format = zathura -P {mark[value]}

Downloaders
===========

.. papis-config:: downloader-proxy

    There is the possibility of download papers using a proxy.
    To know more you can checkout this
    `link <http://docs.python-requests.org/en/master/user/advanced/#proxies>`_.

Databases
=========

.. papis-config:: default-query-string

    This is the default query that a command will take if no
    query string is typed in the command line. For example this is
    the query that is passed to the command open whenever no search
    string is typed:

    ::

        papis open

    Imagine you want to have all your papers whenever you do not
    specify an input query string, then you would set

    ::

        default-query-string = author="John Smith"

    and whenever you typed ``papis open``, onlye the ``John Smith`` authored
    papers would appear. Notice that the current example has been
    done assuming the ``database-backend = papis``.

.. papis-config:: database-backend

    The backend to use in the database. As for now papis supports
    the own database system ``papis`` and
    `whoosh <https://whoosh.readthedocs.io/en/latest/>`_.

.. papis-config:: use-cache

    Set to ``False`` if you do not want to use the ``cache``
    for the given library. This is only effective if you're using the
    ``papis`` database-backend.

.. papis-config:: cache-dir

.. papis-config:: whoosh-schema-fields

    Python list with the ``TEXT`` fields that should be included in the
    whoosh database schema. For instance say that you want to be able
    to search for the ``doi`` and ``ref`` of the documents, then you could
    include

    ::

        whoosh-schema-fields = ['doi', 'ref']

.. papis-config:: whoosh-schema-prototype

    This is the model for the whoosh schema, check
    `the documentation <https://whoosh.readthedocs.io/en/latest/schema.html/>`_
    for more information.

Other
=====

.. papis-config:: citation-string

    string that can be displayed in header if the reference has a
    citation

    Default set to '*'

.. papis-config:: unique-document-keys

    Whenever you add a new document, papis tries to figure out if
    you have already added this document before. This is partially done
    checking for some special keys, and checking if they match.
    Which keys are checked against is decided by this option, which
    should be formatted as a python list, just as in the default value.

    For instance, if you add a paper with a given ``doi``, and then you
    add another document with the same ``doi``, then papis will notify
    you that there is already another document with this ``doi`` because
    the ``doi`` key is part of the ``unique-document-keys`` option.

"""
import logging

logger = logging.getLogger("config")
logger.debug("importing")

import os
import configparser
import papis.exceptions


_CONFIGURATION = None  #: Global configuration object variable.
_DEFAULT_SETTINGS = None  #: Default settings for the whole papis.
_OVERRIDE_VARS = {
    "folder": None,
    "cache": None,
    "file": None,
    "scripts": None
}


def get_default_opener():
    """Get the default file opener for the current system
    """
    import sys
    if sys.platform.startswith('darwin'):
        return "open"
    elif os.name == 'nt':
        return "start"
    elif os.name == 'posix':
        return "xdg-open"


general_settings = {
    "local-config-file": ".papis.config",
    "database-backend": "papis",
    "default-query-string": ".",

    "opentool": get_default_opener(),
    "dir-umask": 0o755,
    "browser": os.environ.get('BROWSER') or get_default_opener(),
    "picktool": "papis.pick",
    "mvtool": "mv",
    "editor": os.environ.get('EDITOR')
                        or os.environ.get('VISUAL')
                        or get_default_opener(),
    "xeditor": get_default_opener(),
    "notes-name": "notes.tex",
    "use-cache": True,
    "cache-dir": None,
    "use-git": False,

    "add-confirm": False,
    "add-name": "",
    "file-name": None,
    "add-interactive": False,
    "add-edit": False,
    "add-open": False,

    "browse-key": 'url',
    "browse-query-format": "{doc[title]} {doc[author]}",
    "search-engine": "https://duckduckgo.com",
    "user-agent": 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3)',
    "scripts-short-help-regex": ".*papis-short-help: *(.*)",
    "info-name": "info.yaml",
    "doc-url-key-name": "doc_url",

    "open-mark": False,
    "mark-key-name": "marks",
    "mark-format-name": "mark",
    "mark-header-format": "{mark[name]} - {mark[value]}",
    "mark-match-format": "{mark[name]} - {mark[value]}",
    "mark-opener-format": get_default_opener(),

    "file-browser": get_default_opener(),
    "bibtex-journal-key": 'journal',
    "extra-bibtex-keys": "",
    "extra-bibtex-types": "",
    "default-library": "papers",
    "export-text-format":
        "{doc[author]}. {doc[title]}. {doc[journal]} {doc[pages]}"
        " {doc[month]} {doc[year]}",
    "format-doc-name": "doc",
    "match-format":
        "{doc[tags]}{doc.subfolder}{doc[title]}{doc[author]}{doc[year]}",
    "format-jinja2-enable": False,
    "header-format-file": None,
    "header-format": \
        "<red>{doc.html_escape[title]}</red>\n"\
        " <span color='#00ff00'>{doc.html_escape[author]}</span>\n"
        "  <span color='#00ffaa'>({doc.html_escape[year]})</span> "
        "[<yellow>{doc.html_escape[tags]}</yellow>]",

    "info-allow-unicode": True,
    "ref-format": "{doc[doi]}",
    "multiple-authors-separator": " and ",
    "multiple-authors-format": "{au[surname]}, {au[given_name]}",

    "whoosh-schema-fields": "['doi']",
    "whoosh-schema-prototype":
    '{\n'
    '"author": TEXT(stored=True),\n'
    '"title": TEXT(stored=True),\n'
    '"year": TEXT(stored=True),\n'
    '"tags": TEXT(stored=True),\n'
    '}',

    "citation-string": "*",
    'unique-document-keys': "['doi','ref','isbn','isbn10','url','doc_url']",

    "tui-editmode": "emacs",
    "downloader-proxy": None,
    "bibtex-unicode": False,

}


def get_general_settings_name():
    """Get the section name of the general settings
    :returns: Section's name
    :rtype:  str
    >>> get_general_settings_name()
    'settings'
    """
    return "settings"


def get_default_settings(section="", key=""):
    """Get the default settings for all non-user variables
    in papis.

    If section and key are given, then the setting
    for the given section and the given key are returned.

    If only ``key`` is given, then the setting
    for the ``general`` section is returned.

    :param section: Particular section of the default settings
    :type  section: str
    :param key: Setting's name to be queried for.
    :type  key: str

    >>> import collections
    >>> type(get_default_settings()) is collections.OrderedDict
    True
    >>> get_default_settings(key='mvtool')
    'mv'
    >>> get_default_settings(key='mvtool', section='settings')
    'mv'
    >>> get_default_settings(key='help-key', section='vim-gui')
    'h'
    """
    global _DEFAULT_SETTINGS
    import papis.gui
    # We use an OrderedDict so that the first entry will always be the general
    # settings, also good for automatic documentation
    from collections import OrderedDict
    if _DEFAULT_SETTINGS is None:
        _DEFAULT_SETTINGS = OrderedDict()
        _DEFAULT_SETTINGS.update({
            get_general_settings_name(): general_settings,
        })
        _DEFAULT_SETTINGS.update(
            papis.gui.get_default_settings()
        )
    if not section and not key:
        return _DEFAULT_SETTINGS
    elif not section:
        return _DEFAULT_SETTINGS[get_general_settings_name()][key]
    else:
        return _DEFAULT_SETTINGS[section][key]


def register_default_settings(settings_dictionary):
    """Register configuration settings into the global configuration registry.

    Notice that you can define sections or global options. For instance,
    let us suppose that a script called ``hubation`` defines some
    configuration options. In the script there could be the following defined

    ::

        import papis.config
        options = {'hubation': { 'command': 'open'}}
        papis.config.register_default_settings(options)

    and later on the script can use these options as:

    ::

        papis.config.get('command', section='hubation')

    :param settings_dictionary: A dictionary with settings
    :type  settings_dictionary: dict
    >>> papis.config.register_default_settings(\
            {'scihub': { 'command': 'open'}}\
        )
    >>> papis.config.get('command', section='scihub')
    'open'
    >>> options = {'settings': { 'hubhub': 42, 'default-library': 'mag' }}
    >>> papis.config.register_default_settings(options)
    >>> papis.config.get('hubhub')
    42
    >>> papis.config.get('info-name') is not None
    True
    >>> not papis.config.get('default-library') == 'mag'
    True
    >>> papis.config.get_default_settings(key='default-library') == 'mag'
    True
    """
    default_settings = get_default_settings()
    # we do a for loop because apparently the OrderedDict removes all
    # key-val fields after updating, so we have to do it by hand
    for section in settings_dictionary.keys():
        if section in default_settings.keys():
            default_settings[section].update(settings_dictionary[section])
        else:
            default_settings[section] = settings_dictionary[section]


def get_config_home():
    """Returns the base directory relative to which user specific configuration
    files should be stored.

    :returns: Configuration base directory
    :rtype:  str
    """
    xdg_home = os.environ.get('XDG_CONFIG_HOME')
    if xdg_home:
        return os.path.expanduser(xdg_home)
    else:
        return os.path.join(os.path.expanduser('~'), '.config')


def get_config_dirs():
    """Get papis configuration directories where the configuration
    files might be stored
    """
    dirs = []
    if os.environ.get('XDG_CONFIG_DIRS'):
        # get_config_home should also be included on top of XDG_CONFIG_DIRS
        dirs += [
            os.path.join(d, 'papis') for d in
            os.environ.get('XDG_CONFIG_DIRS').split(':')
        ]
    # Take XDG_CONFIG_HOME and $HOME/.papis for backwards
    # compatibility
    dirs += [
        os.path.join(get_config_home(), 'papis'),
        os.path.join(os.path.expanduser('~'), '.papis')
    ]
    return dirs


def get_config_folder():
    """Get folder where the configuration files are stored,
    e.g. ``/home/user/.papis``. It is XDG compatible, which means that if the
    environment variable ``XDG_CONFIG_HOME`` is defined it will use the
    configuration folder ``XDG_CONFIG_HOME/papis`` instead.

    """
    config_dirs = get_config_dirs()
    for config_dir in config_dirs:
        if os.path.exists(config_dir):
            return config_dir
    # If no folder is found, then get the config home
    return os.path.join(get_config_home(), 'papis')


def get_config_file():
    """Get the path of the main configuration file,
    e.g. /home/user/.papis/config
    """
    global _OVERRIDE_VARS
    if _OVERRIDE_VARS["file"] is not None:
        config_file = _OVERRIDE_VARS["file"]
    else:
        config_file = os.path.join(
            get_config_folder(), "config"
        )
    logger.debug("Getting config file %s" % config_file)
    return config_file


def set_config_file(filepath):
    """Override the main configuration file path
    """
    global _OVERRIDE_VARS
    logger.debug("Setting config file to %s" % filepath)
    _OVERRIDE_VARS["file"] = filepath


def get_scripts_folder():
    """Get folder where the scripts are stored,
    e.g. /home/user/.papis/scripts
    """
    return os.path.join(
        get_config_folder(), "scripts"
    )


def set(key, val, section=None):
    """Set a key to val in some section and make these changes available
    everywhere.
    """
    config = get_configuration()
    if not config.has_section(section or "settings"):
        config.add_section(section or "settings")
    # FIXME: Right now we can only save val in string form
    # FIXME: It would be nice to be able to save also int and booleans
    # FIXME: configparser seems to be able to only store string values
    config.set(section or get_general_settings_name(), key, str(val))


def general_get(key, section=None, data_type=None):
    """General getter method that will be specialized for different modules.

    :param data_type: The data type that should be expected for the value of
        the variable.
    :type  data_type: DataType, e.g. int, src ...
    :param default: Default value for the configuration variable if it is not
        set.
    :type  default: It should be the same that ``data_type``
    :param extras: List of tuples containing section and prefixes
    """
    # Init main variables
    method = None
    value = None
    config = get_configuration()
    lib = get_lib()
    global_section = get_general_settings_name()
    specialized_key = section + "-" + key if section is not None else key
    extras = [(section, key)] if section is not None else []
    sections = [(global_section, specialized_key)] +\
        extras + [(lib, specialized_key)]
    default_settings = get_default_settings()

    # Check data type for setting getter method
    if data_type == int:
        method = config.getint
    elif data_type == float:
        method = config.getfloat
    elif data_type == bool:
        method = config.getboolean
    else:
        method = config.get

    # Try to get key's value from configuration
    for extra in sections:
        sec = extra[0]
        whole_key = extra[1]
        if sec not in config.keys():
            continue
        if whole_key in config[sec].keys():
            value = method(sec, whole_key)

    if value is None:
        try:
            default = default_settings[
                section or global_section
            ][
                specialized_key if section is None else key
            ]
        except KeyError as e:
            raise papis.exceptions.DefaultSettingValueMissing(key)
        else:
            return default
    return value


def get(*args, **kwargs):
    """String getter
    """
    return general_get(*args, **kwargs)


def getint(*args, **kwargs):
    """Integer getter
    >>> set('something', 42)
    >>> getint('something')
    42
    """
    return general_get(*args, data_type=int, **kwargs)


def getfloat(*args, **kwargs):
    """Float getter
    >>> set('something', 0.42)
    >>> getfloat('something')
    0.42
    """
    return general_get(*args, data_type=float, **kwargs)


def getboolean(*args, **kwargs):
    """Bool getter
    >>> set('add-open', True)
    >>> getboolean('add-open')
    True
    """
    return general_get(*args, data_type=bool, **kwargs)


def get_configuration():
    """Get the configuration object, if no papis configuration has ever been
    initialized, it initializes one. Only one configuration per process should
    ever be configured.

    :returns: Configuration object
    :rtype:  papis.config.Configuration
    """
    global _CONFIGURATION
    if _CONFIGURATION is None:
        logger.debug("Creating configuration")
        _CONFIGURATION = Configuration()
        # Handle local configuration file, and merge it if it exists
        local_config_file = papis.config.get("local-config-file")
        merge_configuration_from_path(local_config_file, _CONFIGURATION)
    return _CONFIGURATION


def merge_configuration_from_path(path, configuration):
    """
    Merge information of a configuration file found in `path`
    to the information of the configuration object stored in `configuration`.

    :param path: Path to the configuration file
    :type  path: str
    :param configuration: Configuration object
    :type  configuration: papis.config.Configuration
    """
    logger.debug("Merging configuration from " + path)
    configuration.read(path)
    configuration.handle_includes()


def set_lib(library):
    """Set library, notice that in principle library can be a full path.

    :param library: Library name or path to a papis library
    :type  library: str

    """
    config = get_configuration()
    if library not in config.keys():
        if os.path.exists(library):
            # Check if the path exists, then use this path as a new library
            logger.debug("Using library %s" % library)
            config[library] = dict(dir=library)
        else:
            raise Exception(
                "Path or library '%s' does not seem to exist" % library
            )
    os.environ["PAPIS_LIB"] = library
    os.environ["PAPIS_LIB_DIR"] = get('dir')


def get_lib():
    """Get current library, it either retrieves the library from
    the environment PAPIS_LIB variable or from the command line
    args passed by the user.

    :param library: Name of library or path to a given library
    :type  library: str
    """
    try:
        lib = os.environ["PAPIS_LIB"]
    except KeyError:
        # Do not put papis.config.get because get is a special function
        # that also needs the library to see if some key was overridden!
        lib = papis.config.get_default_settings(key="default-library")
    return lib


def reset_configuration():
    """Destroys existing configuration and returns a new one.

    :returns: Configuration object
    :rtype:  papis.config.Configuration
    """
    global _CONFIGURATION
    if _CONFIGURATION is not None:
        logger.warning("Overwriting previous configuration")
    _CONFIGURATION = None
    logger.debug("Resetting configuration")
    return get_configuration()


class Configuration(configparser.ConfigParser):

    default_info = {
      "papers": {
        'dir': '~/Documents/papers'
      },
      get_general_settings_name(): {
        'default-library': 'papers'
      }
    }

    logger = logging.getLogger("Configuration")

    def __init__(self):
        configparser.ConfigParser.__init__(self)
        self.dir_location = get_config_folder()
        self.scripts_location = get_scripts_folder()
        self.file_location = get_config_file()
        self.initialize()

    def handle_includes(self):
        if "include" in self.keys():
            for name in self["include"]:
                self.logger.debug("including %s" % name)
                self.read(os.path.expanduser(self.get("include", name)))

    def initialize(self):
        if not os.path.exists(self.dir_location):
            self.logger.warning(
                'Creating configuration folder in %s' % self.dir_location
            )
            os.makedirs(self.dir_location)
        if not os.path.exists(self.scripts_location):
            os.makedirs(self.scripts_location)
        if os.path.exists(self.file_location):
            self.read(self.file_location)
            self.handle_includes()
        else:
            for section in self.default_info:
                self[section] = {}
                for field in self.default_info[section]:
                    self[section][field] = self.default_info[section][field]
            with open(self.file_location, "w") as configfile:
                self.write(configfile)



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="iopscience"
        )
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url):
        url = re.sub(r'/pdf', '', url)
        if re.match(r".*iopscience.iop.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_doi(self):
        # http://iopscience.iop.org/article/10.1088/0305-4470/24/2/004?blah=1
        mdoi = re.match(r'.*\.org/[^/]+/([^?]*)', self.get_url())
        if mdoi:
            doi = mdoi.group(1).replace("abs/", "").replace("full/", "")
            self.logger.debug("[doi] = %s" % doi)
            return doi

    def get_document_url(self):
        # http://iopscience.iop.org/article/10.1088/0305-4470/24/2/004/pdf
        durl = 'https://iopscience.iop.org/article/{0}/pdf'.format(
            self.get_doi()
        )
        self.logger.debug("[doc url] = %s" % durl)
        return durl

    def _get_article_id(self):
        """Get article's id for IOP
        :returns: Article id
        """
        doi = self.get_doi()
        if doi:
            articleId = doi.replace('10.1088/', '')
            self.logger.debug("[doc articleId] = %s" % articleId)
            return articleId

    def get_bibtex_url(self):
        # http://iopscience.iop.org/export?
        # articleId=0305-4470/24/2/004&exportFormat=
        # iopexport_bib&exportType=abs&navsubmit=Export%2Babstract
        articleId = self._get_article_id()
        url = "{0}{1}{2}".format(
            "http://iopscience.iop.org/export?articleId=",
            articleId,
            "&exportFormat=iopexport_bib&exportType=abs"
            "&navsubmit=Export%2Babstract"
        )
        self.logger.debug("[bibtex url] = %s" % url)
        return url



import sys

try:
    import dmenu
except:
    print('You don\'t have dmenu for python, try pip3 install dmenu')
    sys.exit(1)

import papis.api
import papis.utils
import functools


_dmenu_pick = functools.partial(
    dmenu.show,
    case_insensitive=papis.config.getboolean(
        'case_insensitive', section="dmenu-gui"
    ),
    lines=papis.config.getint('lines', section='dmenu-gui'),
    bottom=papis.config.getboolean('bottom', section='dmenu-gui'),
    font=papis.config.get('font', section='dmenu-gui'),
    background=papis.config.get('background', section='dmenu-gui'),
    foreground=papis.config.get('foreground', section='dmenu-gui'),
    background_selected=papis.config.get(
        'background_selected', section='dmenu-gui'
    ),
    foreground_selected=papis.config.get(
        'foreground_selected', section='dmenu-gui'
    )
)


def pick(options, header_filter=None, body_filter=None, match_filter=None):
    if header_filter is None:
        fmt = papis.config.get('header-format', section='dmenu-gui')
        if fmt is None:
            fmt = papis.config.get('header-format')

        def header_filter(x):
            return papis.utils.format_doc(fmt, x)
    if len(options) == 1:
        index = 0
    else:
        headers = [header_filter(o) for o in options]
        header = _dmenu_pick(headers)
        if not header:
            return None
        index = headers.index(header)
    return options[index]



import logging

import tempfile
import papis.bibtex
import papis.config

logger = logging.getLogger("downloader")


def get_available_downloaders():
    names = [
        "aps", "acs", "arxiv", "ieee", "scitationaip", "annualreviews",
        "iopscience", "libgen", "get", "thesesfr", "worldscientific",
    ]
    downloaders = []
    for name in names:
        mod = __import__(
            'papis.downloaders.' + name,
            fromlist=['Downloader']
        )
        downloaders.append(getattr(mod, 'Downloader'))
    return downloaders


def get_downloader(url, downloader=False):
    """Get downloader object. If only a url is given, the url is matched
    against the match method of the downloaders.

    :param url: Url of the document
    :type  url: str
    :param downloader: Name of a downloader
    :type  downloader: str
    :returns: A Downloader if found or none
    :rtype:  papis.downloader.base.Downloader

    """
    if not downloader:
        for downloader in get_available_downloaders():
            result = downloader.match(url)
            if result:
                return result
    else:
        try:
            mod = __import__(
                'papis.downloaders.%s' % downloader,
                fromlist=['Downloader']
            )
            return getattr(mod, 'Downloader')(url)
        except ImportError:
            logger.error("No downloader named %s" % downloader)


def get(url, data_format="bibtex", expected_doc_format=None):

    result = {
        "data": dict(),
        "doi": None,
        "documents_paths": []
    }

    downloader = get_downloader(url)
    if not downloader:
        logger.warning(
            "No matching downloader found (%s)" % url
        )
        return result

    logger.info("Using downloader '%s'" % downloader)
    if downloader.expected_document_extension is None and \
            expected_doc_format is not None:
        downloader.expected_document_extension = expected_doc_format

    if data_format == "bibtex":
        try:
            bibtex_data = downloader.get_bibtex_data()
            if bibtex_data:
                result["data"] = papis.bibtex.bibtex_to_dict(
                    bibtex_data
                )
                result["data"] = result["data"][0] \
                    if len(result["data"]) > 0 else dict()
        except NotImplementedError:
            pass

    try:
        doc_data = downloader.get_document_data()
    except NotImplementedError:
        doc_data = None

    try:
        result['doi'] = downloader.get_doi()
    except NotImplementedError:
        pass

    if doc_data is not None:
        if downloader.check_document_format():
            result["documents_paths"].append(tempfile.mktemp())
            logger.info(
                "Saving downloaded data in %s" % result["documents_paths"][-1]
            )
            with open(result["documents_paths"][-1], "wb+") as fd:
                fd.write(doc_data)

    return result



"""
This command is useful if your library is itself a git repository.
You can use this command to issue git commands in your library
repository without having to change your current directory.

CLI Examples
^^^^^^^^^^^^

    - Check the status of the library repository:

    .. code::

        papis git status

    - Commit all changes:

    .. code::

        papis git commit -a


"""
import papis.commands
import papis.commands.run
import papis.config
import logging
import click

logger = logging.getLogger('git')


def run(folder, command=[]):
    return papis.commands.run.run(folder, command=["git"] + command)


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.help_option('--help', '-h')
@click.argument("command", nargs=-1)
def cli(command):
    "Run a git command in the library folder"
    return run(papis.config.get("dir"), command=list(command))



"""This command will try its best to find a source in the internet for the
document at hand.

Of course if the document has an url key in its info file, it will use this url
to open it in a browser.  Also if it has a doc_url key, or a doi, it will try
to compose urls out of these to open it.

If none of the above work, then it will try to use a search engine with the
document's information (using the ``browse-query-format``).  You can select
wich search engine you want to use using the ``search-engine`` setting.

Cli
^^^
.. click:: papis.commands.browse:cli
    :prog: papis browse
"""
import papis
import papis.utils
import papis.config
from papis.api import status
import papis.cli
import papis.api
import click
import papis.database


def run(document):
    """Open document's url in a browser"""
    papis.document.open_in_browser(document)
    return status.success


@click.command()
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    '-k', '--key', default='',
    help='Use the value of the document\'s key to open in the browser, e.g.'
         'doi, url, doc_url ...'
)
def cli(query, key):
    """Open document's url in a browser"""
    documents = papis.database.get().query(query)
    document = papis.api.pick_doc(documents)
    if len(key):
        papis.config.set('browse-key', key)
    if not document:
        return status.file_not_found
    return run(document)



"""

Examples
^^^^^^^^

- To override some configuration options, you can use the flag ``--set``, for
  instance, if you want to override the editor used and the opentool to open
  documents, you can just type

    .. code:: shell

        papis --set editor gedit --set opentool firefox edit
        papis --set editor gedit --set opentool firefox open

- If you want to list the libraries and pick one before sending a database
  query to papis, use ``--pick-lib`` as such

    .. code:: shell

        papis --pick-lib open 'einstein relativity'

Cli
^^^
.. click:: papis.commands.default:run
    :prog: papis
    :commands: []

"""
import os
import papis
import papis.api
import papis.config
import papis.commands
import logging
import click
import papis.cli


@click.group(
    cls=papis.cli.MultiCommand,
    invoke_without_command=True
)
@click.help_option('--help', '-h')
@click.version_option(version=papis.__version__)
@click.option(
    "-v",
    "--verbose",
    help="Make the output verbose (equivalent to --log DEBUG)",
    default=False,
    is_flag=True
)
@click.option(
    "-l",
    "--lib",
    help="Choose a library name or library path (unamed library)",
    default=lambda: papis.config.get("default-library")
)
@click.option(
    "-c",
    "--config",
    help="Configuration file to use",
    default=None,
)
@click.option(
    "--log",
    help="Logging level",
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO"
)
@click.option(
    "--pick-lib",
    help="Pick library to use",
    default=False,
    is_flag=True
)
@click.option(
    "--cc", "--clear-cache", "clear_cache",
    help="Clear cache of the library used",
    default=False,
    is_flag=True
)
@click.option(
    "-s", "--set", "set_list",
    type=(str, str),
    multiple=True,
    help="Set key value, e.g., "
         "--set info-name information.yaml  --set opentool evince",
)
def run(
        verbose,
        config,
        lib,
        log,
        pick_lib,
        clear_cache,
        set_list
        ):
    log_format = '%(levelname)s:%(name)s:%(message)s'
    if verbose:
        log = "DEBUG"
        log_format = '%(relativeCreated)d-'+log_format
    logging.basicConfig(
        level=getattr(logging, log),
        format=log_format
    )
    logger = logging.getLogger('default')

    for pair in set_list:
        logger.debug('Setting "{0}" to "{1}"'.format(*pair))
        papis.config.set(pair[0], pair[1])

    if config:
        papis.config.set_config_file(config)
        papis.config.reset_configuration()

    if pick_lib:
        lib = papis.api.pick(
            papis.api.get_libraries(),
            pick_config=dict(header_filter=lambda x: x)
        )

    papis.config.set_lib(lib)

    # Now the library should be set, let us check if there is a
    # local configuration file there, and if there is one, then
    # merge its contents
    local_config_file = os.path.expanduser(
        os.path.join(
            papis.config.get("dir"),
            papis.config.get("local-config-file")
        )
    )
    papis.config.merge_configuration_from_path(
        local_config_file,
        papis.config.get_configuration()
    )

    if clear_cache:
        papis.api.clear_lib_cache(lib)



import logging
import urllib.request  # urlopen, Request
import urllib.parse  # import urlencode
import papis.config
import json


logger = logging.getLogger('dissemin')


def dissemin_authors_to_papis_authors(data):
    new_data = dict()
    if 'authors' in data.keys():
        authors = []
        for author in data['authors']:
            keys = ('first', 'last')
            authors.append(
                dict(
                    given_name=author['name']['first'],
                    surname=author['name']['last']
                )
            )
        new_data["author_list"] = authors
        new_data["author"] = ",".join(
            ["{a[given_name]} {a[surname]}".format(a=a) for a in authors]
        )
    return new_data


def dissemindoc_to_papis(data):
    common_data = dict()
    result = []
    common_data['title'] = data.get('title')
    common_data['type'] = data.get('type')
    common_data.update(dissemin_authors_to_papis_authors(data))
    for record in data['records']:
        new_data = dict()
        new_data.update(common_data)
        new_data.update(record)
        new_data['doc_url'] = new_data.get('pdf_url')
        new_data['url'] = new_data.get('splash_url')
        new_data['tags'] = new_data.get('keywords')

        new_data = {key: new_data[key] for key in new_data if new_data[key]}
        result.append(new_data)
    return result


def get_data(query=None):
    """
    Get data using the dissemin API
    https://dissem.in/api/search/?q=pregroup
    """
    dict_params = { "q": query, }
    result = []
    clean_params = {x: dict_params[x] for x in dict_params if dict_params[x]}
    params = urllib.parse.urlencode(dict_params)
    main_url = "https://dissem.in/api/search/?"
    req_url = main_url + params
    logger.debug("url = " + req_url)
    url = urllib.request.Request(
        req_url,
        headers={
            'User-Agent': papis.config.get('user-agent')
        }
    )
    jsondoc = urllib.request.urlopen(url).read().decode()
    paperlist = json.loads(jsondoc)
    docs = sum([dissemindoc_to_papis(d) for d in paperlist['papers']], [])
    return docs



import papis.config
import logging


class DocMatcher(object):
    """This class implements the mini query language for papis.
    All its methods are static, it could be also implemented as a separate
    module.

    The static methods are to be used as follows:
    First the search string has to be set,
        DocMatcher.set_search(search_string)
    and then the parse method should be called in order to decypher the
    search_string,
        DocMatcher.parse()
    Now the DocMatcher is ready to match documents with the input query
    via the `return_if_match` method, which is used to parallelize the
    matching.
    """
    search = ""
    parsed_search = None
    if papis.config.get('format-jinja2-enable'):
        doc_format = '{{' + papis.config.get('format-doc-name') + '["DOC_KEY"]}}' 
    else:
        doc_format = '{' + papis.config.get('format-doc-name') + '[DOC_KEY]}'
    logger = logging.getLogger('DocMatcher')
    matcher = None

    @classmethod
    def return_if_match(cls, doc):
        """Use the attribute `cls.parsed_search` to match the `doc` document
        to the previously parsed query.
        :param doc: Papis document to match against.
        :type  doc: papis.document.Document
        :returns: True if it matches, False if some query requirement does
            not match.

        >>> import papis.document
        >>> from papis.database.cache import match_document
        >>> doc = papis.document.from_data(dict(title='einstein'))
        >>> DocMatcher.set_matcher(match_document)
        >>> DocMatcher.parse('einste')
        ([(['einste'], {})], {})
        >>> DocMatcher.return_if_match(doc) is not None
        True
        >>> DocMatcher.parse('heisenberg')
        ([(['heisenberg'], {})], {})
        >>> DocMatcher.return_if_match(doc) is not None
        False
        >>> DocMatcher.parse('title = ein')
        ([(['title', '=', 'ein'], {})], {})
        >>> DocMatcher.return_if_match(doc) is not None
        True

        """
        match = None
        for parsed in cls.parsed_search:
            if len(parsed) == 1:
                search = parsed[0]
                sformat = None
            elif len(parsed) == 3:
                search = parsed[2]
                sformat = cls.doc_format.replace('DOC_KEY', parsed[0])
            match = doc if cls.matcher(doc, search, sformat) else None
            if not match:
                break
        return match

    @classmethod
    def set_search(cls, search):
        """
        >>> DocMatcher.set_search('author = Hummel')
        >>> DocMatcher.search
        'author = Hummel'
        """
        cls.search = search

    @classmethod
    def set_matcher(cls, matcher):
        """
        >>> from papis.database.cache import match_document
        >>> DocMatcher.set_matcher(match_document)
        """
        cls.matcher = matcher

    @classmethod
    def parse(cls, search=None):
        """Parse the main query text. This method will also set the
        class attribute `parsed_search` to the parsed query, and it will
        return it too.
        :param cls: The class object, since it is a static method
        :type  cls: object
        :param search: Search text string if a custom search string is to be
            used. False if the `cls.search` class attribute is to be used.
        :type  search: str
        :returns: Parsed query
        :rtype:  list
        >>> print(DocMatcher.parse('hello author = einstein'))
        [['hello'], ['author', '=', 'einstein']]
        >>> print(DocMatcher.parse(''))
        []
        >>> print(\
            DocMatcher.parse(\
                '"hello world whatever =" tags = \\\'hello ====\\\''))
        [['hello world whatever ='], ['tags', '=', 'hello ====']]
        >>> print(DocMatcher.parse('hello'))
        [['hello']]
        """
        if search is None:
            search = cls.search
        cls.parsed_search = parse_query(search)
        return cls.parsed_search


def parse_query(query_string):
    """
    >>> print(parse_query('hello   author = einstein'))
    [['hello'], ['author', '=', 'einstein']]
    """
    import pyparsing
    logger = logging.getLogger('query_parser')
    logger.debug('Parsing search')
    papis_key = pyparsing.Word(pyparsing.alphanums + '-')
    papis_value = pyparsing.QuotedString(
        quoteChar='"', escChar='\\', escQuote='\\'
    ) ^ pyparsing.QuotedString(
        quoteChar="'", escChar='\\', escQuote='\\'
    ) ^ papis_key
    equal = pyparsing.ZeroOrMore(" ") + \
        pyparsing.Literal('=') + \
        pyparsing.ZeroOrMore(" ")

    papis_query = pyparsing.ZeroOrMore(
        pyparsing.Group(
            pyparsing.ZeroOrMore(
                papis_key + equal
            ) + papis_value
        )
    )
    parsed = papis_query.parseString(query_string)
    logger.debug('Parsed query = %s' % parsed)
    return parsed



"""
The export command is useful to work with other programs such as bibtex.

Some examples of its usage are:

- Export one of the documents matching the author with einstein to bibtex:

.. code::

    papis export --bibtex 'author = einstein'

or export all of them

.. code::

    papis export --bibtex --all 'author = einstein'

- Export all documents to bibtex and save them into a ``lib.bib`` file

.. code::

    papis export --all --bibtex --out lib.bib

- Export a folder of one of the documents matching the word ``krebs``
  into a folder named, ``interesting-document``

.. code::

    papis export --folder --out interesting-document krebs

  this will create the folder ``interesting-document`` containing the
  ``info.yaml`` file, the linked documents and a ``bibtex`` file for
  sharing with other people.

  You can also just export its associated document:

.. code::

    papis export --file krebs


Cli
^^^
.. click:: papis.commands.export:cli
    :prog: papis export
"""
import papis
import os
import sys
import shutil
import papis.utils
import papis.document
import click
import papis.cli
import papis.api
import papis.database
import logging


def run(
    documents,
    yaml=False,
    bibtex=False,
    json=False,
    text=False
):
    """
    Exports several documents into something else.

    :param document: A ist of papis document
    :type  document: [papis.document.Document]
    :param yaml: Wether to return a yaml string
    :type  yaml: bool
    :param bibtex: Wether to return a bibtex string
    :type  bibtex: bool
    :param json: Wether to return a json string
    :type  json: bool
    :param text: Wether to return a text string representing the document
    :type  text: bool
    """
    if json:
        import json
        return json.dumps(
            [
                papis.document.to_dict(document) for document in documents
            ]
        )

    if yaml:
        import yaml
        return yaml.dump_all(
            [
                papis.document.to_dict(document) for document in documents
            ],
            allow_unicode=True
        )

    if bibtex:
        return '\n'.join([
            papis.document.to_bibtex(document) for document in documents
        ])

    if text:
        text_format = papis.config.get('export-text-format')
        return '\n'.join([
            papis.utils.format_doc(text_format, document)
            for document in documents
        ])

    return None


@click.command()
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    "--yaml",
    help="Export into yaml",
    default=False,
    is_flag=True
)
@click.option(
    "--bibtex",
    help="Export into bibtex",
    default=False,
    is_flag=True
)
@click.option(
    "--json",
    help="Export into json",
    default=False,
    is_flag=True
)
@click.option(
    "--folder",
    help="Export document folder to share",
    default=False,
    is_flag=True
)
@click.option(
    "-o",
    "--out",
    help="Outfile or outdir",
    default=None
)
@click.option(
    "-t",
    "--text",
    help="Text formated reference",
    default=False,
    is_flag=True
)
@click.option(
    "-a", "--all",
    help="Export all without picking",
    default=False,
    is_flag=True
)
@click.option(
    "--file",
    help="Export a copy of a file",
    default=False,
    is_flag=True
)
def cli(
        query,
        yaml,
        bibtex,
        json,
        folder,
        out,
        text,
        all,
        file
        ):
    """Export a document from a given library"""

    logger = logging.getLogger('cli:export')
    documents = papis.database.get().query(query)

    if json and folder or yaml and folder:
        logger.warning("Only --folder flag will be considered")

    if not all:
        document = papis.api.pick_doc(documents)
        if not document:
            return 0
        documents = [document]

    ret_string = run(
        documents,
        yaml=yaml,
        bibtex=bibtex,
        json=json,
        text=text
    )

    if ret_string is not None:
        if out is not None:
            logger.info("Dumping to {0}".format(out))
            with open(out, 'w+') as fd:
                fd.write(ret_string)
        else:
            logger.info("Dumping to stdout")
            print(ret_string)
        return 0

    for document in documents:
        if folder:
            folder = document.get_main_folder()
            outdir = out or document.get_main_folder_name()
            if not len(documents) == 1:
                outdir = os.path.join(
                    outdir, document.get_main_folder_name()
                )
            shutil.copytree(folder, outdir)
        elif file:
            logger.info("Exporting file")
            files = document.get_files()
            assert(isinstance(files, list))
            if not files:
                logger.error('No files found for doc in {0}'.format(
                    document.get_main_folder()
                ))
                continue
            files_to_open = [papis.api.pick(
                files,
                pick_config=dict(
                    header_filter=lambda x: x.replace(
                        document.get_main_folder(), ""
                    )
                )
            )]
            files_to_copy = list(filter(lambda x: x, files_to_open))
            for file_to_open in files_to_copy:

                if out is not None and len(files_to_open) == 1:
                    out_file = out
                else:
                    out_file = os.path.basename(file_to_open)

                logger.info("copy {0} to {1}".format(
                    file_to_open,
                    out_file
                ))
                shutil.copyfile(
                    file_to_open,
                    out_file
                )



import re
import urllib.parse
import urllib.request
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="ieee")
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url):
        m = re.match(r"^ieee:(.*)", url, re.IGNORECASE)
        if m:
            url = "http://ieeexplore.ieee.org/document/{m}".format(
                m=m.group(1)
            )
        if re.match(r".*ieee.org.*", url):
            # http://http://ieeexplore.ieee.org/document/7989161/
            url = re.sub(r"\.pdf.*$", "", url)
            return Downloader(url)
        else:
            return False

    def get_identifier(self):
        url = self.get_url()
        return re.sub(r'^.*ieeexplore.ieee.org/document/(.*)\/', r'\1', url)

    def get_bibtex_url(self):
        identifier = self.get_identifier()
        bibtex_url = \
            'http://ieeexplore.ieee.org/xpl/downloadCitations?reload=true'
        data = {
            'recordIds': identifier,
            'citations-format': 'citation-and-abstract',
            'download-format': 'download-bibtex',
            'x': '0',
            'y': '0'
        }
        return bibtex_url, data

    def download_bibtex(self):
        bib_url, values = self.get_bibtex_url()
        post_data = urllib.parse.urlencode(values)
        post_data = post_data.encode('ascii')

        self.logger.debug("[bibtex url] = %s" % bib_url)

        req = urllib.request.Request(bib_url, post_data)
        with urllib.request.urlopen(req) as response:
            data = response.read()
            text = data.decode('utf-8')
            text = text.replace('<br>', '')
            self.bibtex_data = text

    def get_document_url(self):
        identifier = self.get_identifier()
        self.logger.debug("[paper id] = %s" % identifier)
        pdf_url = "{}{}".format(
            "http://ieeexplore.ieee.org/"
            "stampPDF/getPDF.jsp?tp=&isnumber=&arnumber=",
            identifier
        )
        self.logger.debug("[pdf url] = %s" % pdf_url)
        return pdf_url



import os
import re
from prompt_toolkit.formatted_text.html import HTML, html_escape
from prompt_toolkit.application import (
    Application as PromptToolkitApplication,
)
from prompt_toolkit.history import FileHistory
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.screen import Point
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.widgets import (
    HorizontalLine
)
import papis.config
import logging
logger = logging.getLogger('pick')

__all__ = ['Picker', 'pick']


def create_keybindings(picker):
    kb = KeyBindings()

    @kb.add('c-q')
    @kb.add('c-c')
    def exit_(event):
        picker.deselect()
        event.app.exit()

    @kb.add('c-n')
    @kb.add('down')
    def down_(event):
        picker.options_list.move_down()
        picker.refresh()

    @kb.add('c-p')
    @kb.add('up')
    def up_(event):
        picker.options_list.move_up()
        picker.refresh()

    @kb.add('end')
    def up_(event):
        picker.options_list.go_bottom()
        picker.refresh()

    @kb.add('c-g')
    @kb.add('home')
    def up_(event):
        picker.options_list.go_top()
        picker.refresh()

    @kb.add('c-y')
    @kb.add('c-up')
    @kb.add('s-up')
    def up_(event):
        picker.scroll_up()
        picker.refresh_prompt()

    @kb.add('c-e')
    @kb.add('c-down')
    @kb.add('s-down')
    def up_(event):
        picker.scroll_down()
        picker.refresh_prompt()

    @kb.add('f1')
    def help(event):
        picker.options_list.content.text = """
Bindings:

Ctrl-e, Ctrl-down, Shift-down : Scroll Down
Ctrl-y, Ctrl-up,   Shift-up   : Scroll up
Ctrl-n, down                  : Next item
Ctrl-p, up                    : Previous item
Ctrl-q, Ctrl-c                : Quit
Home                          : First item
End                           : Last item

"""
        picker.refresh_prompt()

    # @kb.add('/')
    # def up_(event):
        # picker.layout.focus(picker.search_buffer)

    # @kb.add('c-f')
    # def up_(event):
        # if picker.layout.has_focus(picker.content_window):
            # picker.layout.focus(picker.search_buffer)
        # else:
            # picker.layout.focus(picker.content_window)

    @kb.add('enter')
    def enter_(event):
        event.app.exit()

    return kb


class OptionsListControl:

    def __init__(
            self,
            options,
            search_buffer,
            default_index=0,
            header_filter=None,
            match_filter=None
            ):

        self.options = options
        self.search_buffer = search_buffer
        self.header_filter = header_filter
        self.match_filter = match_filter
        self.current_index = default_index
        self.entries_left_offset = 2

        self.process_options()

        self.content = FormattedTextControl(
            key_bindings=None,
            #get_cursor_position=self.index_to_point,
            get_cursor_position=lambda: self.cursor,
            show_cursor=True,
            text=''
        )
        self.cursor = Point(0,0)

    def set_options(self, options):
        self.options = options
        self.process_options()

    def index_to_point(self):
        try:
            index = self.indices.index(self.current_index)
            line = sum(
                self.options_headers_linecount[i]
                for i in self.indices[0:index]
            )
            self.cursor = Point(0, line)
        except Exception as e:
            self.cursor = Point(0,0)

    def move_up(self):
        try:
            index = self.indices.index(self.current_index)
            index -= 1
            if index < 0:
                self.current_index = self.indices[-1]
            else:
                self.current_index = self.indices[index]
        except ValueError:
            pass

    def move_down(self):
        try:
            index = self.indices.index(self.current_index)
            index += 1
            if index >= len(self.indices):
                self.current_index = self.indices[0]
            else:
                self.current_index = self.indices[index]
        except ValueError:
            pass

    def go_top(self):
        if len(self.indices) > 0:
            self.current_index = self.indices[0]

    def go_bottom(self):
        if len(self.indices) > 0:
            self.current_index = self.indices[-1]

    def get_search_regex(self):
        cleaned_search = self.search_buffer.text.replace('(', '\\(')\
                                       .replace(')', '\\)')\
                                       .replace('+', '\\+')\
                                       .replace('[', '\\[')\
                                       .replace(']', '\\]')
        return r".*"+re.sub(r"\s+", ".*", cleaned_search)

    def update(self, *args):
        self.filter_options()
        self.refresh()

    def filter_options(self, *args):
        indices = []
        regex = self.get_search_regex()
        for index, option in enumerate(list(self.options)):
            if re.match(regex, self.options_matchers[index], re.I):
                indices += [index]
        self.indices = indices
        if len(self.indices) and self.current_index not in self.indices:
            if self.current_index < min(self.indices):
                self.current_index = self.indices[0]
            elif self.current_index > max(self.indices):
                self.current_index = max(self.indices)
            else:
                self.current_index = self.indices[0]

    def get_selection(self):
        if len(self.indices) and self.current_index is not None:
            return self.options[self.current_index]

    def refresh(self):
        self.index_to_point()
        i = self.current_index
        oldtuple = self.options_headers[i][0]
        self.options_headers[i][0] = (
            oldtuple[0],
            '>' + re.sub(r'^ ', '', oldtuple[1]),
        )
        self.content.text = sum(
            [self.options_headers[i] for i in self.indices],
            []
        )
        self.options_headers[i][0] = oldtuple

    def process_options(self):
        logger.debug('processing options')
        self.options_headers_linecount = [
            len(self.header_filter(o).split('\n'))
            for o in self.options
        ]
        logger.debug('processing headers')
        self.options_headers = []
        for o in self.options:
            prestring = re.sub(
                r'^', ' ' * self.entries_left_offset,
                re.sub(
                    r'\n', '\n' + ' ' * self.entries_left_offset,
                    self.header_filter(o)
                )
            ) + '\n'
            try:
                htmlobject = HTML(prestring).formatted_text
            except:
                logger.error(
                    'Error processing html for \n {0}'.format(prestring)
                )
                htmlobject = HTML(html_escape(prestring)).formatted_text
            self.options_headers += [htmlobject]
        logger.debug('processing matchers')
        self.options_matchers = [self.match_filter(o) for o in self.options]
        self.indices = range(len(self.options))
        logger.debug('options processed')


class Picker(object):
    """The :class:`Picker <Picker>` object

    :param options: a list of options to choose from
    :param title: (optional) a title above options list
    :param indicator: (optional) custom the selection indicator
    :param default_index: (optional) set this if the default
        selected option is not the first one
    """

    def __init__(
            self,
            options,
            title=None,
            indicator='*',
            default_index=0,
            header_filter=lambda x: x,
            match_filter=lambda x: x
            ):

        self.title = title
        self.indicator = indicator
        self.search = ""

        search_buffer_history = FileHistory(
            os.path.join('.', 'search_history')
        )

        self.search_buffer = Buffer(
            history=search_buffer_history,
            enable_history_search=True,
            multiline=False
        )

        self.prompt_buffer = Buffer(multiline=False)


        self.options_list = OptionsListControl(
            options,
            self.search_buffer,
            default_index,
            header_filter,
            match_filter
        )

        self.search_buffer.on_text_changed += self.update

        self.content_window = Window(
            content=self.options_list.content,
            height=None,
            wrap_lines=False,
            ignore_content_height=True,
            always_hide_cursor=True,
            allow_scroll_beyond_bottom=True,
        )
        root_container = HSplit([
            Window(height=1, content=BufferControl(buffer=self.search_buffer)),
            HorizontalLine(),
            self.content_window,
            Window(height=1, content=BufferControl(buffer=self.prompt_buffer)),
        ])

        self.layout = Layout(root_container)

        self.application = self._create_application()
        self.update()

    def deselect(self):
        self.options_list.current_index = None

    def _create_application(self):
        return PromptToolkitApplication(
            input=None,
            output=None,
            editing_mode=EditingMode.EMACS
            if papis.config.get('tui-editmode') == 'emacs'
            else EditingMode.VI,
            layout=self.layout,
            key_bindings=create_keybindings(self),
            include_default_pygments_style=False,
            full_screen=True,
            enable_page_navigation_bindings=True
        )

    def refresh_prompt(self):
        self.prompt_echo(
            "{0}/{1}  F1:help".format(
                int(self.options_list.current_index) + 1,
                len(self.options_list.options),
            )
        )

        # self.prompt_echo(
            # "{0}/{1} ={2}-{3} +{4} -{5}  <{6},{7}>".format(
                # int(self.options_list.current_index) + 1,
                # len(self.options_list.options),
                # self.displayed_lines[0] if self.displayed_lines is not None \
                    # else '',
                # self.displayed_lines[-1] if self.displayed_lines is not None \
                    # else '',
                # self.options_list.cursor,
                # self.content_height,
                # self.first_visible_line,
                # self.last_visible_line,
            # )
        # )

    def scroll_down(self):
        lvl = self.last_visible_line
        ll = self.content_height
        if ll and lvl:
            if lvl + 1 < ll:
                new = lvl + 1
            else:
                new = lvl
            self.options_list.cursor = Point(0, new)

    def scroll_up(self):
        fvl = self.first_visible_line
        if fvl:
            if fvl >= 0:
                new = fvl - 1
            else:
                new = 0
            self.options_list.cursor = Point(0, new)

    def scroll_up(self):
        dp = self.displayed_lines
        if len(dp):
            self.options_list.cursor = Point(0, dp[0] - 1)

    def refresh(self, *args):
        self.options_list.refresh()
        self.refresh_prompt()

    def update(self, *args):
        self.options_list.update()
        self.refresh_prompt()

    @property
    def screen_height(self):
        return self.options_list.content.preferred_height(None, None, None)

    @property
    def displayed_lines(self):
        info = self.content_window.render_info
        if info:
            return info.displayed_lines

    @property
    def first_visible_line(self):
        info = self.content_window.render_info
        if info:
            return info.first_visible_line()

    @property
    def last_visible_line(self):
        info = self.content_window.render_info
        if info:
            return info.last_visible_line()

    @property
    def content_height(self):
        info = self.content_window.render_info
        if info:
            return info.content_height

    def prompt_echo(self, text):
        self.prompt_buffer.text = text

    def run(self):
        if len(self.options_list.options) == 0:
            return ""
        if len(self.options_list.options) == 1:
            return self.options_list.options[0]
        return self.application.run()


def pick(
        options,
        title="Pick: ",
        indicator='>',
        default_index=0,
        header_filter=lambda x: x,
        match_filter=lambda x: x
        ):
    """Construct and start a :class:`Picker <Picker>`.
    """

    if len(options) == 0:
        return ""
    if len(options) == 1:
        return options[0]

    picker = Picker(
                options,
                title,
                indicator,
                default_index,
                header_filter,
                match_filter
                )
    picker.run()
    return picker.options_list.get_selection()



"""
Vim gui
*******

.. papis-config:: help-key
    :section: vim-gui

.. papis-config:: open-key
    :section: vim-gui

.. papis-config:: edit-key
    :section: vim-gui

.. papis-config:: search-key
    :section: vim-gui

.. papis-config:: delete-key
    :section: vim-gui

.. papis-config:: open-dir-key
    :section: vim-gui

.. papis-config:: next-search-key
    :section: vim-gui

.. papis-config:: prev-search-key
    :section: vim-gui

.. papis-config:: header-format
    :section: vim-gui

dmenu gui
*********

See `dmenu <https://tools.suckless.org/dmenu/>`_ and the python wrapper
`here <http://dmenu.readthedocs.io/en/latest/>`_ for more information.
You will need to install the latter to make use of this function,

::

    pip3 install dmenu


.. papis-config:: lines
    :section: dmenu-gui

.. papis-config:: case_insensitive
    :section: dmenu-gui

.. papis-config:: bottom
    :section: dmenu-gui

.. papis-config:: font
    :section: dmenu-gui

.. papis-config:: background
    :section: dmenu-gui

.. papis-config:: foreground
    :section: dmenu-gui

.. papis-config:: background_selected
    :section: dmenu-gui

.. papis-config:: foreground_selected
    :section: dmenu-gui

.. papis-config:: header-format
    :section: dmenu-gui

    This is not set per default, and it will default to
    the general header-format if not set.

"""


def get_default_settings():
    return {
        "vim-gui": {
            "help-key": "h",
            "open-key": "o",
            "edit-key": "e",
            "search-key": "/",
            "delete-key": "dd",
            "open-dir-key": "<S-o>",
            "next-search-key": "n",
            "prev-search-key": "N",
            "header-format":
                "Title : {doc[title]}\n"
                "Author: {doc[author]}\n"
                "Year  : {doc[year]}\n"
                "-------\n",
        },
        "dmenu-gui": {
            "lines": 20,
            "case_insensitive": True,
            "bottom": True,
            "font": 'monospace-14',
            "background": '#000000',
            "foreground": '#55ff55',
            "background_selected": '#005500',
            "foreground_selected": '#f0f0f0',
            "header-format": None,
        },
    }



"""
The open command is a very important command in the papis workflow.
With it you can open documents, folders or marks.

Marks
^^^^^

One of special things about this command is the possibility of
creating marks for documents. As you would imagine, it is in general
difficult to create marks for any kind of data. For instance,
if our library consists of pdf files and epub files for instance,
we would like to define bookmarks in order to go back to them at
some later point.

How you define marks can be customized through the marks configuration
settings :ref:`here <marks-options>`.
The default way of doing it is just by defining a ``marks`` list in a document.
Let us look at a concrete example:

.. code:: yaml

    author: Isaiah Shavitt, Rodney J. Bartlett
    edition: '1'
    files: [book.pdf]
    isbn: 052181832X,9780521818322

    marks:
    - {name: Intermediates definition, value: 344}
    - {name: EOM equations, value: 455}

    publisher: Cambridge University Press
    ref: book:293288
    series: Cambridge Molecular Science
    title: 'Many-Body Methods in Chemistry and Physics'
    type: book
    year: '2009'

This book has defined two marks. Each mark has a name and a value.
If you tell the open command to open marks, then it will look for
the marks and open the value (page number). This is the default behaviour,
however if you go to the :ref:`configuration <marks-options>`
you'll see that you can change the convention to what it suits you.


Examples
^^^^^^^^
- Open a pdf file linked to a document matching the string ``bohm``

    ::

        papis open bohm

- Open the folder where this last document is stored

    ::

        papis open -d bohm

  Please notice that the file browser used will be also related to
  the :ref:`file-browser setting <config-settings-file-browser>`.

- Open a mark defined in the info file

    ::

        papis open --mark bohm


Cli
^^^
.. click:: papis.commands.open:cli
    :prog: papis open
"""
import papis
import papis.api
import papis.utils
import papis.config
import papis.cli
import papis.database
import click
import logging


def run(document, opener=None, folder=False, mark=False):
    logger = logging.getLogger('open:run')
    if opener is not None:
        papis.config.set("opentool", opener)

    if folder:
        # Open directory
        papis.api.open_dir(document.get_main_folder())
    else:
        if mark:
            logger.debug("Getting document's marks")
            marks = document[papis.config.get("mark-key-name")]
            if marks:
                logger.debug("Picking marks")
                mark = papis.api.pick(
                    marks,
                    dict(
                        header_filter=lambda x: papis.utils.format_doc(
                            papis.config.get("mark-header-format"),
                            x, key=papis.config.get("mark-format-name")
                        ),
                        match_filter=lambda x: papis.utils.format_doc(
                            papis.config.get("mark-header-format"),
                            x, key=papis.config.get("mark-format-name")
                        )
                    )
                )
                if mark:
                    opener = papis.utils.format_doc(
                        papis.config.get("mark-opener-format"),
                        mark, key=papis.config.get("mark-format-name")
                    )
                    papis.config.set("opentool", opener)
        files = document.get_files()
        if len(files) == 0:
            logger.error("The document chosen has no files attached")
            return 1
        file_to_open = papis.api.pick(
            files,
            pick_config=dict(
                header_filter=lambda x: x.replace(
                    document.get_main_folder(), ""
                )
            )
        )
        papis.api.open_file(file_to_open, wait=False)


@click.command()
@click.help_option('-h', '--help')
@papis.cli.query_option()
@click.option(
    "--tool",
    help="Tool for opening the file (opentool)",
    default=""
)
@click.option(
    "-d",
    "--dir",
    help="Open directory",
    default=False,
    is_flag=True
)
@click.option(
    "--all",
    help="Open all matching documents",
    default=False,
    is_flag=True
)
@click.option(
    "-m",
    "--mark/--no-mark",
    help="Open mark",
    default=lambda: True if papis.config.get('open-mark') else False
)
def cli(query, tool, dir, all, mark):
    """Open document from a given library"""
    if tool:
        papis.config.set("opentool", tool)

    documents = papis.database.get().query(query)
    if not documents:
        click.echo("No documents found with that name.")
        return 1

    if not all:
        documents = [papis.api.pick_doc(documents)]
        documents = [d for d in documents if d]

    for document in documents:
        run(
            document,
            folder=dir,
            mark=mark
        )



"""

Cli
^^^
.. click:: papis.commands.rename:cli
    :prog: papis rename
"""
import papis
import os
import papis.api
import papis.utils
import subprocess
import logging
import click
import papis.cli
import papis.database


def run(document, new_name, git=False):
    db = papis.database.get()
    logger = logging.getLogger('rename:run')
    folder = document.get_main_folder()
    subfolder = os.path.dirname(folder)

    new_folder_path = os.path.join(subfolder, new_name)

    if os.path.exists(new_folder_path):
        logger.warning("Path %s already exists" % new_folder_path)
        return 1

    cmd = ['git', '-C', folder] if git else []
    cmd += ['mv', folder, new_folder_path]

    logger.debug(cmd)
    subprocess.call(cmd)

    if git:
        papis.utils.git_commit(message="Rename %s" % folder)

    db.delete(document)
    logger.debug("New document folder: {}".format(new_folder_path))
    document.set_folder(new_folder_path)
    db.add(document)
    return 0


@click.command()
@click.help_option('--help', '-h')
@papis.cli.query_option()
@papis.cli.git_option()
def cli(query, git):
    """Rename entry"""

    documents = papis.database.get().query(query)
    document = papis.api.pick_doc(documents)
    if not document:
        return 0

    new_name = papis.utils.input(
        "Enter new folder name:\n"
        ">",
        default=document.get_main_folder_name()
    )
    return run(document, new_name, git=git)



import logging
import isbnlib
# See https://github.com/xlcnd/isbnlib for details

logger = logging.getLogger('papis:isbnlib')


def get_data(query="", service=None):
    global logger
    results = []
    logger.debug('Trying to retrieve isbn')
    isbn = isbnlib.isbn_from_words(query)
    data = isbnlib.meta(isbn, service=service)
    if data is None:
        return results
    else:
        logger.debug('Trying to retrieve isbn')
        assert(isinstance(data, dict))
        results.append(data_to_papis(data))
        return results


def data_to_papis(data):
    """Convert data from isbnlib into papis formated data

    :param data: Dictionary with data
    :type  data: dict
    :returns: Dictionary with papis keynames

    """
    data = {k.lower(): data[k] for k in data}
    return data



import re
import papis.downloaders.base
import bs4


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="thesesfr")
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url):
        if re.match(r".*theses.fr.*", url):
            return Downloader(url)
        else:
            return False

    def get_identifier(self):
        """
        >>> d = Downloader("http://www.theses.fr/2014TOU30305")
        >>> d.get_identifier()
        '2014TOU30305'
        >>> d = Downloader("http://www.theses.fr/2014TOU30305.bib/?asdf=2")
        >>> d.get_identifier()
        '2014TOU30305'
        """
        m = re.match(r".*theses.fr/([^/?.&]+).*", self.url)
        return m.group(1) if m is not None else None

    def get_document_url(self):
        """
        >>> d = Downloader("http://www.theses.fr/2014TOU30305")
        >>> d.get_document_url()
        'http://thesesups.ups-tlse.fr/2722/1/2014TOU30305.pdf'
        >>> d = Downloader("http://theses.fr/1998ENPC9815")
        >>> d.get_document_url()
        """
        raw_data = self.session.get(self.url).content.decode('utf-8')
        soup = bs4.BeautifulSoup(raw_data, "html.parser")
        a = list(filter(
            lambda t: re.match(r'.*en ligne.*', t.text),
            soup.find_all('a')
        ))

        if not a:
            self.logger.error('No document url in theses.fr')
            return None

        second_url = a[0]['href']
        raw_data = self.session.get(second_url).content.decode('utf-8')
        soup = bs4.BeautifulSoup(raw_data, "html.parser")
        a = list(filter(
            lambda t: re.match(r'.*pdf$', t['href']),
            soup.find_all('a')
        ))

        if not a:
            self.logger.error('No document url in {0}'.format(second_url))
            return None

        return a[0]['href']

    def get_bibtex_url(self):
        """
        >>> d = Downloader("http://www.theses.fr/2014TOU30305")
        >>> d.get_bibtex_url()
        'http://www.theses.fr/2014TOU30305.bib'
        """
        url = "http://www.theses.fr/{id}.bib".format(id=self.get_identifier())
        self.logger.debug("[bibtex url] = %s" % url)
        return url

